

from dataclasses import dataclass
from typing import Any
from datetime import datetime
from bioblend.toolshed import ToolShedInstance

from galaxy2janis.gx.gxtool.requirements import CondaRequirement, ContainerRequirement, Requirement
from galaxy2janis.gx.wrappers import Wrapper
from galaxy2janis.runtime.dates import JANIS_DATE_FMT
from galaxy2janis.runtime.dates import TOOLSHED_DATE_FMT


def request_installable_revision(gxstep: dict[str, Any]) -> list[Wrapper]:
    api_interactor = interactor_factory(gxstep)
    print(f"making galaxy API revision requests for {api_interactor.tool_id} v{api_interactor.tool_build}")
    return get_installable_revision(api_interactor)


@dataclass
class ToolshedAPIInteractor:
    tool_shed: str
    owner: str
    repo: str
    tool_id: str
    tool_build: str

    # TODO MULTIPLE REQUEST ATTEMPTS WITH TIMEOUT???

    def set_revision(self, revision: str) -> None:
        self.revision = revision

    def get_installable_revision(self) -> list[Wrapper]:
        ts = ToolShedInstance(f"https://{self.tool_shed}/")
        response = ts.repositories.get_repository_revision_install_info(self.repo, self.owner, self.revision) # type: ignore
        if response and len(response[0]) > 0:  # type: ignore
            return galaxy_response_to_wrappers(response) # type: ignore
        return []

    def get_ordered_revisions(self) -> list[str]:
        ts = ToolShedInstance(f"https://{self.tool_shed}/")
        return ts.repositories.get_ordered_installable_revisions(self.repo, self.owner) # type: ignore


def galaxy_response_to_wrappers(data: list[dict[str, Any]]) -> list[Wrapper]:
    wrappers: list[Wrapper] = []
    repository, metadata, _ = data
    for tool in metadata['valid_tools']:
        details = {
            'owner': repository['owner'],
            'repo': repository['name'],
            'revision': parse_revision(data),
            'tool_id': tool['id'],
            'tool_build': tool['version'],
            'date_created': parse_date_created(data),
            'requirements': parse_requirements(tool),
        }
        wrappers.append(Wrapper(details))
    return wrappers

def parse_requirements(tool: dict[str, Any]) -> list[Requirement]:
    out: list[Requirement] = []
    for req in tool['requirements']:
        if req['type'] == 'package':
            out.append(CondaRequirement(req['name'], req['version']))
        else:
            raise RuntimeError('got something other than conda requirement')
    return out

def parse_revision(data: list[dict[str, Any]]) -> str:
    install_info = data[2]
    assert(len(install_info)) == 1
    return list(install_info.values())[0][2]

def parse_date_created(data: list[dict[str, Any]]) -> str:
    """cast the galaxy toolshed date format to janis format"""
    repository = data[0]
    new_date = datetime.strptime(repository['create_time'], TOOLSHED_DATE_FMT)
    return new_date.strftime(JANIS_DATE_FMT)




def interactor_factory(gxstep: dict[str, Any]) -> ToolshedAPIInteractor:
    return ToolshedAPIInteractor(
        tool_shed = gxstep['tool_shed_repository']['tool_shed'],
        owner = gxstep['tool_shed_repository']['owner'],
        repo = gxstep['tool_shed_repository']['name'],
        tool_id = gxstep['tool_id'].rsplit('/', 2)[-2],
        tool_build = gxstep['tool_version']
    )

def get_installable_revision(api_interactor: ToolshedAPIInteractor) -> list[Wrapper]:
    revisions = api_interactor.get_ordered_revisions()
    for revision in reversed(revisions):
        wrappers = get_revision_wrappers(api_interactor, revision)
        if wrappers and version_match(wrappers, api_interactor):
            return wrappers
    raise RuntimeError

def get_revision_wrappers(api_interactor: ToolshedAPIInteractor, revision: str) -> list[Wrapper]:
    api_interactor.set_revision(revision)
    return api_interactor.get_installable_revision()



def version_match(wrappers: list[Wrapper], api_interactor: ToolshedAPIInteractor) -> bool:
    for wrapper in wrappers:
        if wrapper.tool_id == api_interactor.tool_id and wrapper.tool_build == api_interactor.tool_build:
            return True
    return False

# def get_installable_revision_old(gxstep: dict[str, Any]) -> list[Wrapper]:
#     api_interactor = interactor_factory(gxstep)
#     revision = gxstep['tool_shed_repository']['changeset_revision']
#     data = request_revision_exact(api_interactor, revision)
#     if not data:
#         data = request_revision_general(api_interactor)
#     return wrapper_factory(data)















from datetime import datetime

from galaxy2janis.runtime.dates import JANIS_DATE_FMT
from galaxy2janis.runtime.dates import QUAY_DATE_FMT


"""
yes, this is meant to be written and read to disk. how did you know
"""

class Container: 
    def __init__(self, info: dict[str, str]):
        self.image_type: str = info['image_type']
        self.repo: str = info['repo']
        self.tag: str = info['tag']
        self.url: str = info['url']
        self._timestamp: str = info['_timestamp']

    @property
    def registry_host(self) -> str:
        return self.url.split('/', 1)[0]

    @property
    def tool_name(self) -> str:
        return self.repo

    @property
    def tool_version(self) -> str:
        if self.registry_host == 'quay.io':
            return self.tag.split('--', 1)[0]
        raise NotImplementedError()

    @property
    def timestamp(self) -> str:
        """map the timestamp from different formats to unified output"""
        if self.registry_host == 'quay.io': 
            date = datetime.strptime(self._timestamp, QUAY_DATE_FMT)
        else:
            raise NotImplementedError()
        return date.strftime(JANIS_DATE_FMT)








from galaxy2janis.entities.workflow import Workflow
from galaxy2janis.entities.workflow.step.inputs import (
    ConnectionInputValue, 
    InputValue, 
    StaticInputValue, 
    WorkflowInputInputValue
)


def update_components_optionality(value: InputValue, janis: Workflow) -> None:
    if value.component:  # if linked to tool component
        if isinstance(value, WorkflowInputInputValue):
            if not value.is_runtime:
                w_inp = janis.get_input(query_uuid=value.input_uuid)
                if w_inp.optional and not value.component.optional:
                    value.component.forced_optionality = True
        elif isinstance(value, ConnectionInputValue):
            s_out = janis.get_step_output(query_uuid=value.out_uuid) 
            if s_out.tool_output.optional and not value.component.optional:
                value.component.forced_optionality = True
        elif isinstance(value, StaticInputValue):
            if value.is_none:
                value.component.forced_optionality = True







from dataclasses import dataclass
from uuid import uuid4
from galaxy2janis.gx.command.components import OutputComponent


# TODO could probably go direct to janis

@dataclass
class StepOutput:
    step_uuid: str
    is_wflow_out: bool
    tool_output: OutputComponent

    def __post_init__(self):
        self.uuid = str(uuid4())







from __future__ import annotations
from typing import Optional, Protocol
from galaxy2janis.gx.gxtool.param import Param


# defines what a CommandComponent should look like
class CommandComponent(Protocol):
    uuid: str
    gxparam: Optional[Param]
    forced_optionality: Optional[bool]

    @property
    def name(self) -> str:
        ...

    @property
    def optional(self) -> bool:
        ...

    @property
    def array(self) -> bool:
        ...
    
    @property
    def docstring(self) -> Optional[str]:
        ...







from typing import Any, Optional

from galaxy2janis.gx.gxtool.requirements import Requirement
from galaxy2janis.utils.general import global_align

from .Fetcher import Fetcher
from ..Container import Container
from . import utils


class GA4GHFetcher(Fetcher):
    """uses the GA4GH trs api to get tool containers and versions"""

    def fetch(self, requirement: Requirement) -> list[Container]:
        interactor = GA4GHInteractor()
        tools_list = interactor.request_tool_data(requirement.name)
        # TODO finish refactoring. GA4GH was down. 
        if tools_list:
            tool_data = self.select_tool(tools_list)
            version_data = self.get_version_data(tool_data)
            if version_data:
                return self.response_to_containers()
        return []

    def select_tool(self, tool_data: list[dict[str, Any]]) -> dict[str, Any]:
        raise NotImplementedError()

    def get_version_data(self, tool_data: dict[str, Any]) -> Optional[dict[str, Any]]:
        tool_data = self.most_similar_tool(api_results)
        version = self.most_similar_version(tool_data)
        interactor = GA4GHInteractor()
        return interactor.request_version_data(version['url'])

    def response_to_containers(self, data: dict[str, Any]) -> list[Container]:
        raise NotImplementedError()


    def most_similar_tool(containers: list[Container]) -> dict[str, Any]:
        """returns data for the api tool result most similar to the query tool name"""
        result_similarities: list[SimilarityScore] = []
        query_name = requirement.get_text()
        for tool_data in api_results:
            score = global_align(query_name, tool_data['name'])
            result_similarities.append(SimilarityScore(score, tool_data))
        result_similarities.sort(key = lambda x: x.score, reverse=True)
        return result_similarities[0].obj

class GA4GHInteractor:

    def request_tool_data(self, query_tool: str) -> Optional[list[dict[str, Any]]]:
        """
        make api request which returns dict of search results
        select the correct tool from results & return
        only makes requests which return lists of dicts
        """
        api_uri = self.format_tool_request_url(query_tool)
        return utils.make_api_request(api_uri)

    def request_version_data(self, query_url: str) -> list[dict[str, Any]]:
        return utils.make_api_request(query_url)

    def format_tool_request_url(self, tool_query: str) -> str:
        return f'https://api.biocontainers.pro/ga4gh/trs/v2/tools?name={tool_query}&limit=10&sort_field=id&sort_order=asc'

    







from __future__ import annotations
from typing import TYPE_CHECKING, Tuple

if TYPE_CHECKING:
    from galaxy2janis.entities.workflow import WorkflowStep

from ..TextRender import TextRender


class UntranslatedText(TextRender):
    def __init__(self, entity: WorkflowStep):
        super().__init__()
        self.entity = entity

    @property
    def imports(self) -> list[Tuple[str, str]]:
        return []

    def render(self) -> str:
        step = self.entity
        out_str: str = ''
        if step.preprocessing:
            out_str += '\n# PRE-PROCESSING ---------------------\n'
            out_str += f'{step.preprocessing}\n'
        if step.postprocessing:
            out_str += '\n# POST-PROCESSING ---------------------\n'
            out_str += f'{step.postprocessing}\n'
        return out_str






from __future__ import annotations
from .OutputComponent import OutputComponent
from typing import Optional


class WildcardOutput(OutputComponent):
    def __init__(self):
        super().__init__()
        self.verified: bool = False

    @property
    def name(self) -> str:
        name = ''
        if self.gxparam:
            name = self.gxparam.name
        if not name.startswith('out'):
            name = f'out_{name}'
        return name

    @property
    def optional(self) -> bool:
        # NOTE - janis does not allow optional outputs
        return False
   
    @property
    def docstring(self) -> Optional[str]:
        if self.gxparam:
            return self.gxparam.docstring
        return 'output created during runtime. file is collected from working directory'








from datetime import datetime
from typing import Any, Tuple

from bs4 import Tag

from galaxy2janis.runtime.dates import TOOLSHED_DATE_FMT, JANIS_DATE_FMT
import gx.wrappers.scraping.utils as utils

NEW_REPOS_ONLY = True
THREADS = 10


def scrape_revisions() -> None:
    scraper = RevisionScraper()
    scraper.scrape()


def get_revisions_in_repo(owner: str, repo: str) -> list[dict[str, str]]:
    URL = format_repo_url(owner, repo)
    page = utils.get_soup(URL)
    table = utils.get_table(page)
    ages = utils.get_table_col(table, 0)
    links = utils.get_table_col(table, 2)
    return extract_revisions_info(ages, links)

def format_repo_url(owner: str, repo: str) -> str:
    return f'https://toolshed.g2.bx.psu.edu/repos/{owner}/{repo}/'

def extract_revisions_info(ages: list[Tag], links: list[Tag]) -> list[dict[str, str]]:
    out: list[dict[str, str]] = []
    for age, link in zip(ages, links):
        revision: str = link.find('a').attrs['href'].rsplit('/', 1)[-1] # type: ignore
        date_created = datetime.strptime(age.text, TOOLSHED_DATE_FMT) # type: ignore
        date_string: str = date_created.strftime(JANIS_DATE_FMT)  # TODO does this mess up the time since no UTC offset?
        out.append({
            'revision': revision,
            'date_created': date_string,
        })
    return out


class RevisionScraper:
    """
    coordinates the scraping process
    handles file loading, updating saving
    """

    def __init__(self):
        self.local_data: dict[str, Any] = utils.load_data(utils.REVISION_DATA_PATH)
        self.total_count: int = 0
        self.scraped_count: int = 0

    @property
    def percentage_complete(self) -> float:
        return (self.scraped_count / self.total_count) * 100

    def scrape(self) -> None:
        owners_repos = self.get_scrapable_repos()
        self.total_count = len(owners_repos)
        print(f'repos to scrape: {self.total_count}')
        for owner, repo in owners_repos:
            self.try_scrape_repo_revisions(owner, repo)
            self.scraped_count += 1
            print(f'\r{self.percentage_complete:0.1f}%', end='')

    def get_scrapable_repos(self) -> list[Tuple[str, str]]:
        repo_data = utils.load_data(utils.REPO_DATA_PATH)
        out: list[Tuple[str, str]] = []
        for owner, repo_list in repo_data.items():
            for repo in repo_list:
                if self.should_scrape_repo_revisions(owner, repo):
                    out.append((owner, repo))
        return out

    def should_scrape_repo_revisions(self, owner: str, repo: str) -> bool:
        # if we only want new repos, don't scrape existing repos
        if NEW_REPOS_ONLY:
            if owner in self.local_data:
                if repo in self.local_data[owner]:
                    return False
        return True

    def try_scrape_repo_revisions(self, owner: str, repo: str) -> None:
        if owner not in self.local_data:
            self.local_data[owner] = {}
        try:
            self.local_data[owner][repo] = get_revisions_in_repo(owner, repo)  # takes time - do threaded
            utils.save_data(utils.REVISION_DATA_PATH, self.local_data) # save
        except Exception as e:
            print(e)





from . import patterns

from .matches import get_matches
from .matches import get_next_word
from .matches import get_preceeding_dashes
from .matches import get_quoted_sections
from .matches import find_unquoted

from .checks import is_int
from .checks import is_float
from .checks import is_present
from .checks import is_var
from .checks import is_script
from .checks import has_var







from typing import Any, Optional

from galaxy2janis.runtime.exceptions import AttributeNotSupportedError
from galaxy.tool_util.verify.interactor import ToolTestDescription
from .checks import ValidCheck
from .mapper import map_ttestout

from janis_core.tool.test_classes import (
    TTestCase,
    TTestExpectedOutput
)


"""
for each galaxy test, need to translate into a janis TTestCase. 
inputs are easy, but outputs are more complex. 
for each output, need to specify the TTestExpectedOutput.
this TTestExpectedOutput has a TTestPreprocessor, which obtains the value of the
"""

class TestFactory:
    def produce(self, gxtest: ToolTestDescription) -> TTestCase:
        return TTestCase(
            str(gxtest.name),
            self.prepare_inputs(gxtest),
            self.prepare_outputs(gxtest)
        )

    def prepare_inputs(self, gxtest: ToolTestDescription) -> dict[str, Any]:
        """maps the galaxy test def inputs to janis format. not many changes."""
        inputs: dict[str, Any] = {}
        for key, val in gxtest.inputs.items():
            if isinstance(val, list) and len(val) == 1:
                val = val[0]
            inputs[str(key)] = val
        return inputs
        
    def prepare_outputs(self, gxtest: ToolTestDescription) -> list[TTestExpectedOutput]:
        outputs: list[TTestExpectedOutput] = []
        for gxout in gxtest.outputs:
            checks = self.gather_checks(gxout)
            outputs += [map_ttestout(c) for c in checks]
        return outputs

    def gather_checks(self, gxout: dict[str, Any]) -> list[ValidCheck]:
        checks: list[ValidCheck] = []
        checks += self.gather_diff_checks(gxout)
        checks += self.gather_re_match_checks(gxout)
        checks += self.gather_sim_size_checks(gxout)
        checks += self.gather_file_fingerprint_checks(gxout)
        checks += self.gather_file_assert_checks(gxout)
        checks += self.gather_file_contains_checks(gxout)
        return checks

    def gather_diff_checks(self, gxout: dict[str, Any]) -> list[ValidCheck]:
        if gxout['attributes']['compare'] == 'diff' and gxout['value']:
            return [self.create_check(gxout['name'], 'lines_diff', gxout['attributes']['lines_diff'], reffile=gxout['value'])]
        return []
    
    def gather_re_match_checks(self, gxout: dict[str, Any]) -> list[ValidCheck]:
        if gxout['attributes']['compare'] in ['re_match', 're_match_multiline']:
            raise AttributeNotSupportedError('re_match and re_match_multiline not yet supported')
        return []
    
    def gather_sim_size_checks(self, gxout: dict[str, Any]) -> list[ValidCheck]:
        if gxout['attributes']['compare'] == 'sim_size':
            fields = ['delta', 'delta_frac']
            return self.create_attribute_checks(gxout, fields=fields)
        return []
        
    def gather_file_fingerprint_checks(self, gxout: dict[str, Any]) -> list[ValidCheck]:
        # file fingerprints
        # TODO add ftype checks!
        #fields = ['ftype', 'md5', 'checksum']
        fields = ['md5', 'checksum']

        return self.create_attribute_checks(gxout, fields=fields)

    def gather_file_assert_checks(self, gxout: dict[str, Any]) -> list[ValidCheck]:
        checks: list[ValidCheck] = []
        if gxout['attributes']['assert_list']:
            for assertcheck in gxout['attributes']['assert_list']:
                checks.append(self.create_check(gxout['name'], assertcheck['tag'], assertcheck['attributes']))
        return checks
        
    def gather_file_contains_checks(self, gxout: dict[str, Any]) -> list[ValidCheck]:
        checks: list[ValidCheck] = []
        if gxout['attributes']['compare'] == 'contains' and gxout['value']:
            checks.append(self.create_check(gxout['name'], 'has_text', None, reffile=gxout['value']))
        return checks

    def create_attribute_checks(self, gxout: dict[str, Any], valid_fields: Optional[list[str]] = None) -> list[ValidCheck]:
        """creates an attribute checks based on the values of valid_valid_fields"""
        if not valid_fields:
            valid_fields = []
        return [ self.create_check(gxout['name'], attname, attval) 
                 for attname, attval in gxout['attributes'].items() 
                 if attname in valid_fields and attval]

    def create_check(self, name: str, ctype: str, value: Optional[Any], reffile: Optional[str]=None) -> ValidCheck:
        return ValidCheck(name, ctype, value, reffile)

    def create_check2(self, name: str, ctype: str, value: Optional[Any], reffile: Optional[str]=None) -> ValidCheck:
        if value:
            if not isinstance(value, dict):
                test_values: dict[str, Any] = {'value': value}
            else:
                test_values = value
        else:
            test_values: dict[str, Any] = {}
        return ValidCheck(name, ctype, test_values, reffile)


        





            





    #     # get the right map object 
    #     ttestouts = []
    #     for out in gxtest.outputs:
    #         ttestouts += self.init_file_diff_outputs(out)
    #         ttestouts += self.init_file_check_outputs(out)
    #     return ttestouts

    # def init_file_diff_outputs(self, gx_test_out: dict[str, Any]) -> list[TTestExpectedOutput]:
    #     outs = []
    #     if gx_test_out['value']:
    #         self.init_diff_outputs(gx_test_out)
    #     return outs

    # def init_file_check_outputs(self, gx_test_out: dict[str, Any]) -> list[TTestExpectedOutput]:
    #     outs = []
    #     outs += self.init_fingerprint_outputs(gx_test_out)
    #     outs += self.init_rematch_outputs(gx_test_out)
    #     outs += self.init_simsize_outputs(gx_test_out)
    #     outs += self.init_assert_outputs(gx_test_out)
    #     return outs

    # def init_fingerprint_outputs(self, gx_test_out: dict[str, Any]) -> list[TTestExpectedOutput]:
    #     pass

    # def init_fingerprint_outputs(self, gx_test_out: dict[str, Any]) -> list[TTestExpectedOutput]:
    #     if gx_test_out['attributes']['compare'] in ['re_match', 're_match_multiline']:
    #         raise AttributeNotSupportedError('re_match and re_match_multiline output checks not supported')

    # def init_simsize_outputs(self, gx_test_out: dict[str, Any]) -> list[TTestExpectedOutput]:
    #     outchecks = ['']
    #     outs = []
    #     for check in outchecks:
    #         outs += strategy_map[check](gx_test_out)
    #     return outs

    # def init_assert_outputs(self, gx_test_out: dict[str, Any]) -> list[TTestExpectedOutput]:
    #      if gx_test_out['attributes']['assert_list']:
    #         for out_check in gx_test_out['attributes']['assert_list']:
    #             self.map_assert_output()





"""
COMPARE FILE NEEDED -> may need decompression if decompress="true"
compare="diff":
- lines_diff
- ftype

compare="sim_size":
do filesize check with the following info
- delta
maximum allowed absolute size difference (in bytes) between the data set that is generated in the test and the file in test-data/ that is referenced by the file attribute. Default value is 10000 bytes. Can be combined with delta_frac.
- delta_frac
If compare is set to sim_size, this is the maximum allowed relative size difference between the data set that is generated in the test and the file in test-data/ that is referenced by the file attribute. A value of 0.1 means that the file that is generated in the test can differ by at most 10% of the file in test-data. The default is not to check for relative size difference. Can be combined with delta.


FILE NOT NEEDED
md5
checksum
assert_list has elems



"""





    # def init_output(self):
    #     pass
    #     for output in gxtest.outputs:
    #         self.
    #     if gxtest.outputs
    #     #preprocessor_map = self.
    #     compare_method = gxtest.output_data['compare']
    #     return TTestExpectedOutput(
    #         gxtest.name,
    #         preprocessor_map[compare_method],
    #     )
        
        







from __future__ import annotations
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from galaxy2janis.entities.workflow import WorkflowStep
    from galaxy2janis.entities.workflow import Workflow
    from galaxy2janis.entities.workflow import WorkflowInput

from galaxy2janis.logs import logging

from galaxy2janis.gx.command.cmdstr import gen_command_string
from galaxy2janis.gx.command.components import Flag
from galaxy2janis.gx.command.components import InputComponent
from galaxy2janis.gx.command.components import Option

from galaxy2janis.gx.gxtool import load_xmltool
from galaxy2janis.gx.gxtool.text import load_partial_cheetah_command

from . import factory
from galaxy2janis.entities.workflow import InputValue

from galaxy2janis import mapping
from galaxy2janis import datatypes
from galaxy2janis import settings
from galaxy2janis import expressions


def handle_tool_static_inputs(janis: Workflow, galaxy: dict[str, Any]) -> None:
    """supplied values in step 'tool_state'"""
    for g_step in galaxy['steps'].values():
        if g_step['type'] == 'tool':
            j_step = mapping.step(g_step['id'], janis, galaxy)
            settings.tool.set(wrapper=j_step.metadata.wrapper)
            ingest_values_cheetah(g_step, j_step, janis)
            ingest_values_inputs(g_step, j_step, janis)

def ingest_values_cheetah(g_step: dict[str, Any], j_step: WorkflowStep, janis: Workflow) -> None:
    ingestor = CheetahInputIngestor(g_step, j_step, janis)
    ingestor.ingest()

def ingest_values_inputs(g_step: dict[str, Any], j_step: WorkflowStep, janis: Workflow) -> None:
    ingestor = StaticInputIngestor(g_step, j_step, janis)
    ingestor.ingest()


class CheetahInputIngestor:
    def __init__(self, g_step: dict[str, Any], j_step: WorkflowStep, janis: Workflow):
        self.g_step = g_step
        self.j_step = j_step
        self.janis = janis

    def ingest(self) -> None:
        cmdstr = self.prepare_command()
        for component in self.get_linkable_components():
            match component:
                case Flag():
                    self.link_flag(component, cmdstr)
                case Option():
                    self.link_option(component, cmdstr)
                case _:
                    pass

    def prepare_command(self) -> str:
        xmltool = load_xmltool()
        command = load_partial_cheetah_command(inputs_dict=self.g_step['tool_state'])
        cmdstr = gen_command_string(source='xml', the_string=command, xmltool=xmltool)
        stmtstr = cmdstr.main.cmdline
        logging.runtime_data(command)
        logging.runtime_data(stmtstr)
        return stmtstr
    
    def link_flag(self, flag: Flag, cmdstr: str) -> None:
        """
        links a flag component value as None if not in cmdstr
        should only detect the flag's absense, nothing else
        """
        if not expressions.is_present(flag.prefix, cmdstr):
            self.handle_not_present_flag(flag)

    def link_option(self, option: Option, cmdstr: str) -> None:
        """gets the value for a specific tool argument"""
        value = expressions.get_next_word(option.prefix, option.delim, cmdstr)
        value = None if value == '' else value
        if value is None:
            self.handle_not_present_opt(option)
        elif self.is_param(value):
            self.handle_gxvar_opt(option, value)
        elif expressions.is_var(value) or expressions.has_var(value):
            self.handle_envvar_opt(option, value)
        else:
            self.handle_value_opt(option, value)

    # TODO upgrade for pre/post task section
    def is_param(self, text: Optional[str]) -> bool:
        # how does this even work? 
        # $ can also mean env var? 
        if text:
            if text[0] == '$':
                return True
            elif len(text) > 1 and text[1] == '$':  # WTF 
                return True
        return False

    def handle_not_present_flag(self, flag: Flag) -> None:
        self.update_tool_values_static(component=flag, value=False)

    def handle_not_present_opt(self, option: Option) -> None:
        self.update_tool_values_static(component=option, value=None)
    
    def handle_gxvar_opt(self, option: Option, value: str) -> None:
        # TODO future: attach the identified param if not attached? 
        # should always be attached tho? 
        pass
    
    def handle_envvar_opt(self, option: Option, value: Any) -> None:
        self.update_tool_values_runtime(component=option)
    
    def handle_value_opt(self, option: Option, value: Any) -> None:
        self.update_tool_values_static(component=option, value=value)

    def update_tool_values_static(self, component: Flag | Option, value: Any) -> None:
        # create & add value 
        is_default = True if component.default_value == value else False
        inputval = factory.static(component, value, default=is_default)
        self.j_step.inputs.add(inputval)
    
    def update_tool_values_runtime(self, component: Flag | Option) -> None:
        # create & add new workflow input
        winp = self.create_workflow_input(component)
        self.janis.add_input(winp)
        # create & add value 
        inputval = factory.workflow_input(component, winp.uuid, is_runtime=True)
        self.j_step.inputs.add(inputval)

    def create_workflow_input(self, component: Flag | Option) -> WorkflowInput:
        """creates a workflow input for the tool input component"""
        return WorkflowInput(
            _name=component.tag,
            array=component.array,
            is_runtime=True,
            datatype=datatypes.get(component),
            optional=component.optional
        )

    def get_linkable_components(self) -> list[InputComponent]:
        out: list[InputComponent] = []
        for component in self.j_step.tool.inputs:
            if not self.j_step.inputs.get(component.uuid):
                out.append(component)
        return out



class StaticInputIngestor:
    def __init__(self, g_step: dict[str, Any], j_step: WorkflowStep, janis: Workflow):
        self.g_step = g_step
        self.j_step = j_step
        self.janis = janis

    def ingest(self) -> None:
        for component in self.get_linkable_components():
            if self.is_directly_linkable(component):
                value = self.create_value(component)
                self.j_step.inputs.add(value)
    
    def get_linkable_components(self) -> list[InputComponent]:
        out: list[InputComponent] = []
        # tool components which don't yet appear in register
        tool_inputs = self.j_step.tool.inputs
        tool_values = self.j_step.inputs
        for component in tool_inputs:
            if component.gxparam and not tool_values.get(component.uuid):
                out.append(component)
        return out

    def is_directly_linkable(self, component: InputComponent) -> bool:
        """
        checks whether a janis tool input can actually be linked to a value in the 
        galaxy workflow step.
        only possible if the component has a gxparam, and that gxparam is referenced as a
        ConnectionStepInput, RuntimeStepInput or StaticStepInput
        """
        if component.gxparam:
            query = component.gxparam.name 
            if query in self.g_step['tool_state']:
                return True
        return False

    def create_value(self, component: InputComponent) -> InputValue:
        """
        create an InputValue for this tool component using supplied step 'tool_state' input values.
        we know that the component has a galaxy param, and that the same galaxy param
        has a supplied value (or connection or runtime value specified) in the step input values. 
        this function grabs that supplied value, then creates a formalised InputValue. 
        it also does other necessary functions - in the case of a galaxy 'runtime value', 
        this needs to become a WorkflowInput in janis world. a WorkflowInput would be created, 
        added to the Workflow, and also added to the InputValue (WorkflowInputInputValue). 
        this says 'for step x using tool y, the tool input component z gets its value from 
        our new WorkflowInput'
        """
        # pull value from 'tool_state'
        # should only be static values left
        # this should be really ez?
        
        g_value = self.g_step['tool_state'][component.gxparam.name] # type: ignore
        is_default = True if component.default_value == g_value else False
        return factory.static(component, value=g_value, default=is_default)






from ..gxtool.XMLToolDefinition import XMLToolDefinition
from .factory import CommandFactory
from .Command import Command


def gen_command(xmltool: XMLToolDefinition) -> Command:
    factory = CommandFactory(xmltool)
    return factory.create()









from typing import Any
import json
import unittest
import xml.etree.ElementTree as et

from gx.gxworkflow.parsing.tool_state import load_tool_state
from gx.gxtool.text.cheetah.evaluation import sectional_evaluate


UNICYCLER_VANILLA_PATH = 'test/data/command/manipulation/template/unicycler/unicycler_command.xml'
UNICYCLER_TEMPLATED_PATH = 'test/data/command/manipulation/template/unicycler/unicycler_command_templated.xml'
UNICYCLER_INPUTS_PATH = 'test/data/command/manipulation/template/unicycler/unicycler_step.json'


def read_cmd(path: str) -> str:
    tree = et.parse(path)
    root = tree.getroot()
    assert(root.text)
    return root.text

def read_step_inputs(path: str) -> dict[str, Any]:
    with open(path, 'r') as fp:
        step = json.load(fp)
    step['tool_state'] = load_tool_state(step)
    return step['tool_state']


class TestSectionalCheetah(unittest.TestCase):

    def test_unicycler(self):
        vanilla = read_cmd(UNICYCLER_VANILLA_PATH)
        reference = read_cmd(UNICYCLER_TEMPLATED_PATH)
        inputs = read_step_inputs(UNICYCLER_INPUTS_PATH)
        templated = sectional_evaluate(vanilla, inputs)
        self.assertEquals(reference, templated)


        




# components
from .CommandComponent import CommandComponent
from .inputs.InputComponent import InputComponent
from .inputs.Flag import Flag
from .inputs.Option import Option
from .inputs.Positional import Positional

from .linux.Tee import Tee
from .linux.streams import StreamMerge

from .outputs.OutputComponent import OutputComponent
from .outputs.InputOutput import InputOutput
from .outputs.RedirectOutput import RedirectOutput
from .outputs.WildcardOutput import WildcardOutput

from . import factory




from .JanisDatatype import JanisDatatype
from .core import bool_t
from .core import file_t
from .core import string_t
from .core import directory_t
from .core import float_t
from .core import int_t
from .core import CORE_DATATYPES
from .core import DEFAULT_DATATYPE
from .register import register

core_type_priorities = {
    bool_t.classname: 1,  # highest priority
    directory_t.classname: 2,
    float_t.classname: 3,
    string_t.classname: 4,
    int_t.classname: 5,
    file_t.classname: 6, # lowest priority
}

def galaxy_to_janis(galaxy_types: list[str]) -> list[JanisDatatype]:
    out: list[JanisDatatype] = []
    for gtype in galaxy_types:
        jtype = register.get(gtype)
        if jtype is not None:
            out.append(jtype)
    return out

def janis_to_core(query_types: list[JanisDatatype]) -> list[JanisDatatype]:
    core: dict[str, JanisDatatype] = {} # dict to keep unique
    for qtype in query_types:
        if qtype.classname in CORE_DATATYPES:
            core[qtype.classname] = qtype
    return list(core.values())

def select_primary_core_type(query_types: list[JanisDatatype]) -> JanisDatatype:
    if query_types:
        selected_type = query_types[0]
        for jtype in query_types[1:]:
            if core_type_priorities[jtype.classname] > core_type_priorities[selected_type.classname]:
                selected_type = jtype
        return selected_type
    else:
        return DEFAULT_DATATYPE






from .token import Token, TokenType

from galaxy2janis import expressions
from galaxy2janis.expressions.patterns import ALL


def spawn_end_sentinel() -> Token:
    matches = expressions.get_matches('end', ALL)
    return Token(matches[0], TokenType.END_STATEMENT)

def spawn_kv_linker(delim: str) -> Token:
    matches = expressions.get_matches(delim, ALL)
    return Token(matches[0], TokenType.KV_LINKER)

def spawn_function_call() -> Token:
    matches = expressions.get_matches('__FUNCTION_CALL__', ALL)
    return Token(matches[0], TokenType.FUNCTION_CALL)
    
def spawn_backtick_section() -> Token:
    matches = expressions.get_matches('__BACKTICK_SHELL_STATEMENT__', ALL)
    return Token(matches[0], TokenType.BACKTICK_SHELL_STATEMENT)





from logging import getLogger, config
from typing import Optional

import warnings
import yaml
import sys 

from galaxy2janis import paths
from galaxy2janis import settings

# -------------
# configuration
# -------------



# logging 
def configure_logging() -> None:
    config_path = f'{paths.USER_DATA_DIR}/{paths.LOGGING_CONFIG}'
    with open(config_path, "r") as fp:
        the_dict = yaml.safe_load(fp)
        the_dict['handlers']['janis_log']['filename'] = paths.janis_log()
        the_dict['handlers']['message_log']['filename'] = paths.message_log()
    config.dictConfig(the_dict)

# warnings configuration
def configure_warnings() -> None:
    warnings.filterwarnings("ignore")

# -------
# logging
# -------

# messages
def msg_parsing_tool():
    filename = settings.tool.tool_path.rsplit('/', 1)[-1]
    print(f'parsing tool {filename}')

def msg_parsing_workflow():
    filename = settings.workflow.workflow_path
    print(f'\nparsing workflow {filename}')

def msg_downloading_tool(url: str):
    print(f'downloading wrapper from {url}')


# debug
# (just runtime data and info messages which don't impact program)
def runtime_data(data: str):
    logger = getLogger('tool')
    logger.debug(str(data))

def evaluation_failed():
    logger = getLogger('tool')
    logger.debug('cheetah evaluation failed')


# info
# (things that MAY cause issues or just metric collecting)

def has_preprocessing():
    logger = getLogger('tool')
    logger.info('preprocessing')

def has_postprocessing():
    logger = getLogger('tool')
    logger.info('postprocessing')

def has_backtick_statement():
    logger = getLogger('tool')
    logger.info('backtick statement')

def has_multiline_str():
    logger = getLogger('tool')
    logger.info('multiline string')

def has_repeat():
    logger = getLogger('tool')
    logger.info('repeat param')

def has_cheetah_loop(text: Optional[str]=None):
    logger = getLogger('tool')
    logger.info(f'cheetah loop')

def has_cheetah_function():
    logger = getLogger('tool')
    logger.info('cheetah function')

def no_base_cmd():
    logger = getLogger('tool')
    logger.info('no base cmd')

def has_configfile():
    logger = getLogger('tool')
    logger.info('configfile')

def no_container():
    logger = getLogger('tool')
    logger.info('no container')

def no_ga4gh_data():
    logger = getLogger('tool')
    logger.info('no ga4gh data')

def container_version_mismatch():
    logger = getLogger('tool')
    logger.info('container version mismatch')

def color_param_ignored():
    logger = getLogger('tool')
    logger.info('ignored unsupported param type: color')

def uncertain_output():
    logger = getLogger('tool')
    logger.info('uncertain outputs')

def unlinked_input_connection():
    logger = getLogger('workflow')
    logger.info('unknown input')


# warning
# (things that will probably cause issues and require human editing in the workflow)

def no_inputs():
    logger = getLogger('tool')
    logger.warning('no inputs')

def no_outputs():
    logger = getLogger('tool')
    logger.warning('no outputs')

def zero_length_tag():
    logger = getLogger('tool')
    logger.warning('zero length tag')

# error
# (unsupported features)

# critical
# (program failed - uncaught exceptions)

def tool_exception():
    logger = getLogger('tool')
    logger.critical('exception')
    sys.exit(1)

def no_close_quotation():
    logger = getLogger('tool')
    logger.critical('no closing quotation')
    sys.exit(1)

def workflow_exception():
    logger = getLogger('workflow')
    logger.critical('exception')
    sys.exit(1)





import os
import tempfile
from typing import Optional

from galaxy.tools import Tool as GxTool
from galaxy.tool_util.parser import get_tool_source
from galaxy.tools import create_tool_from_source
from galaxy.model import History

from galaxy2janis.gx.interaction.mock import MockApp, MockObjectStore
from galaxy2janis.utils import galaxy as utils


def get_tool(xml_path: str) -> GxTool:
    app = _get_app()
    tool_source = get_tool_source(xml_path)
    tool = create_tool_from_source(app, tool_source)
    tool.assert_finalized()
    return tool

def get_builtin_tool_path(tool_id: str) -> Optional[str]:
    """returns path to xml file with id='tool_id'"""
    tool_directories = _get_builtin_tool_directories()
    for directory in tool_directories:
        xmlfile = utils.get_xml_by_id(directory, tool_id)
        if xmlfile:
            return f'{directory}/{xmlfile}'
    return None

def _get_builtin_tool_directories() -> list[str]:
    out: list[str] = []
    out += _get_builtin_tools_directories()
    out += _get_datatype_converter_directories()
    return out

def _get_builtin_tools_directories() -> list[str]:
    import galaxy.tools
    tools_folder = str(galaxy.tools.__file__).rsplit('/', 1)[0]
    bundled_folders = os.listdir(f'{tools_folder}/bundled')
    bundled_folders = [f for f in bundled_folders if not f.startswith('__')]
    bundled_folders = [f'{tools_folder}/bundled/{f}' for f in bundled_folders]
    bundled_folders = [f for f in bundled_folders if os.path.isdir(f)]
    return [tools_folder] + bundled_folders

def _get_datatype_converter_directories() -> list[str]:
    import galaxy.datatypes
    datatypes_folder = str(galaxy.datatypes.__file__).rsplit('/', 1)[0]
    converters_folder = f'{datatypes_folder}/converters'
    return [converters_folder]

def _get_app() -> MockApp:
    # basic details
    app = MockApp()
    app.job_search = None
    app.object_store = MockObjectStore()
    # config
    app.config.new_file_path = os.path.join(tempfile.mkdtemp(), "new_files")
    app.config.admin_users = "grace@thebest.com"
    app.config.len_file_path = "moocow"
    # database
    
    app.model.context.add(History())
    app.model.context.flush()
    return app





from galaxy2janis.logs import logging
from galaxy2janis import expressions

from ..gxtool.XMLToolDefinition import XMLToolDefinition
from ..gxtool.param.Param import Param
from ..gxtool.param.InputParam import BoolParam, SelectParam

from .epath.ExecutionPath import ExecutionPath
from .epath.ExecutionPathAnnotator import GreedyExecutionPathAnnotator

from .Command import Command 
from .cmdstr.CommandString import CommandString
from .components.inputs.Flag import Flag
from .components.inputs.Option import Option
from .components import factory



# Utility functions to check aspects of a gxparam argument attribute

def argument_exists(gxparam: Param, xmltool: XMLToolDefinition, command: Command) -> bool:
    if hasattr(gxparam, 'argument') and gxparam.argument is not None:  #type: ignore
        return True
    return False

def argument_format(gxparam: Param, xmltool: XMLToolDefinition, command: Command) -> bool:
    banned_argument_chars = [' ', '/', '\\']
    if any([char in gxparam.argument for char in banned_argument_chars]):  #type: ignore
        return False
    return True

def argument_component_not_exists(gxparam: Param, xmltool: XMLToolDefinition, command: Command) -> bool:
    # check whether the argument is already known as a prefix 
    prefix_components: list[Flag | Option] = []
    prefix_components += command.get_flags()
    prefix_components += command.get_options()
    for component in prefix_components:
        if gxparam.argument == component.prefix: # type: ignore
            return False
    return True

def argument_has_command_presence(gxparam: Param, xmltool: XMLToolDefinition, command: Command) -> bool:
    # prefix appears in the cmd text
    if gxparam.argument in xmltool.raw_command: # type: ignore
        return True
    return False

def select_is_flag(gxparam: SelectParam) -> bool:
    """checks if a galaxy SelectParam is being used as a Flag() tool input"""
    values = gxparam.get_all_values(nonempty=True)
    if len(values) == 1:
        if values[0].startswith('-'):
            return True
    return False  

# def bool_is_option(gxparam: BoolParam, xmltool: XMLToolDefinition) -> bool:
#     """
#     checks if a galaxy BoolParam is being used as a Option() tool input.
#     eg cutadapt: $info_file is a BoolParam, but in the command we see 
#     '--info-file=$info_file'. '--info-file $info_file' would also count as an Option()
#     """
#     if isinstance(gxparam, BoolParam):
#         variable_fmt1: str = f'${gxparam.name}'
#         variable_fmt2: str = f'${{{gxparam.name}}}'
#         if variable_fmt1 in text or variable_fmt2 in text:
#             pass
#         #return True
#     text: str = xmltool.raw_command
#     prefix: str = gxparam.argument

#     pattern = r'(?<=\s|^)' + prefix + r'[:=][\'"]?\${?([\w_.])*?' + gxparam.name + r'}?[\'"]?(?=\s|$)'
#     if expressions.get_matches(text, pattern):
#         return True
#     return False


# Annotator classes

class ArgumentCommandAnnotator:
    def __init__(self, command: Command, xmltool: XMLToolDefinition):
        self.command = command
        self.xmltool = xmltool

    def annotate(self) -> None:
        # add any gxparams which hint they are components (or update the component)
        for gxparam in self.xmltool.inputs.list():
            if self.should_update_command_components(gxparam):
                gxparam = self.refine_argument(gxparam)
                self.update_command_components(gxparam)
    
    def should_update_command_components(self, gxparam: Param) -> bool:
        checks = [
            argument_exists,
            argument_format,
            argument_component_not_exists,
            argument_has_command_presence,
        ]
        for check in checks:
            if not check(gxparam, self.xmltool, self.command):
                return False
        return True

    def refine_argument(self, gxparam: Param) -> Param:
        """
        'argument' attribute of gxparam not always written with 
        correct number of preceeding dashes. this aims to discover
        the correct amount by looking in the <command> section for the
        argument
        """
        old_argument: str = gxparam.argument # type: ignore
        matches = expressions.get_preceeding_dashes(
            search_term=old_argument,
            text=self.xmltool.raw_command
        )
        if matches:
            num_dashes = max(len(dashes) for dashes in matches) 
            gxparam.argument = '-' * num_dashes + old_argument # type: ignore
        return gxparam

    def update_command_components(self, gxparam: Param) -> None:
        """
        gxparam definitely has 'argument' field
        the below are assumptions - galaxy XML can be written any way you want. 
        will need finer tuning / more disgression for use cases later?
        """
        assert(gxparam.argument) # type: ignore
        match gxparam:
            case BoolParam():
                pass
                #self.handle_bool_param(gxparam)
            case SelectParam():
                self.handle_select_param(gxparam)
            case _:
                self.handle_generic_param(gxparam)

    # def handle_bool_param(self, gxparam: BoolParam) -> None:
    #     assert(gxparam.argument)
    #     if bool_is_option(gxparam, self.xmltool):
    #         option = factory.option(prefix=gxparam.argument, gxparam=gxparam)
    #         self.command.update(option) 
    #     else:
    #         flag = factory.flag(prefix=gxparam.argument, gxparam=gxparam)
    #         self.command.update(flag)

    def handle_select_param(self, gxparam: SelectParam) -> None:
        assert(gxparam.argument)
        if select_is_flag(gxparam):
            flag = factory.flag(prefix=gxparam.argument, gxparam=gxparam)
            self.command.update(flag)
        else:
            values = [opt.value for opt in gxparam.options]
            option = factory.option(prefix=gxparam.argument, gxparam=gxparam, values=values)
            self.command.update(option)
        
    def handle_generic_param(self, gxparam: Param) -> None:
        option = factory.option(prefix=gxparam.argument, gxparam=gxparam)
        self.command.update(option) 



class CmdstrCommandAnnotator:

    def __init__(self, command: Command, xmltool: XMLToolDefinition, cmdstrs: list[CommandString]):
        self.command = command
        self.xmltool = xmltool
        self.cmdstrs = cmdstrs
        self.epath_count: int = 0

    def annotate(self) -> None:
        for cmdstr in self.cmdstrs:
            for epath in cmdstr.main.get_execution_paths():
                logging.runtime_data(str(epath))
                self.extract_components(epath)

    def extract_components(self, epath: ExecutionPath) -> None:
        epath.id = self.epath_count
        epath = self.assign_epath_components(epath)
        self.update_command(epath)
        self.epath_count += 1
    
    def assign_epath_components(self, epath: ExecutionPath) -> ExecutionPath:
        annotator = GreedyExecutionPathAnnotator(epath, self.xmltool, self.command)
        return annotator.annotate()

    def update_command(self, epath: ExecutionPath) -> None:
        for component in epath.get_components():
            self.command.update(component)






from galaxy2janis.logs import logging
from typing import Optional

from galaxy2janis.gx.gxtool import XMLToolDefinition
from galaxy2janis.gx.gxtool.param import Param

from . import constructs
from . import utils
from ..epath.utils import is_bool_select
from ..tokens import Token
from ..tokens import tokenize


class RealisedTokens:
    """
    class exists to expose hidden values within galaxy params
    each galaxy select param can hold any number of text values,
    and each text value may have 1 or more words / keyval pairs
    """
    def __init__(self, values: list[list[Token]], original: Token):
        self.tlists = values
        self.original = original
        
    def get_original_token(self) -> Token:
        return self.original
    
    def get_default_token(self) -> Token:
        for tlist in self.tlists:
            if len(tlist) > 0:
                return tlist[0]
        return self.get_original_token()
    
    def get_first_word(self) -> str:
        return self.get_default_token().text

    def get_gx_reference(self) -> Optional[Param]:
        # every token will have the same gx object
        return self.get_original_token().gxparam
    
    def __repr__(self) -> str:
        strvalues: list[str] = []
        for tlist in self.tlists:
            strvalues.append(' '.join([token.text for token in tlist]))
        return f'RealisedTokenValues: {", ".join(strvalues)}'


class RealisedTokenFactory:
    def __init__(self, xmltool: Optional[XMLToolDefinition]):
        self.xmltool = xmltool
        self.tracker = constructs.ConstructTracker()  # this is all a bit ugly

    def try_tokenify(self, the_string: str) -> list[RealisedTokens]:
        rtvs: list[RealisedTokens] = []
        try:
            for line in utils.split_lines(the_string):
                self.tracker.update(line)
                if self.should_tokenify_line(line):
                    rtvs += self.tokenify_line(line)
        except ValueError:
            logging.no_close_quotation()
        return rtvs

    def should_tokenify_line(self, line: str) -> bool:
        if self.tracker.active_is_boundary(line) or self.tracker.within_banned_segment:
            return False
        return True
    
    def tokenify_line(self, line: str) -> list[RealisedTokens]:
        line_tokens = self.create_line_tokens(line)
        line_tokens = self.set_token_context(line_tokens)
        return self.create_realised_values(line_tokens)

    def create_line_tokens(self, line: str) -> list[Token]:
        line_tokens: list[Token] = []
        for word in utils.split_to_words(line):
            line_tokens += tokenize(word, xmltool=self.xmltool)
        return line_tokens

    def set_token_context(self, line_tokens: list[Token]) -> list[Token]:
        for token in line_tokens:
            token.construct = self.tracker.stack.current_construct
            token.in_conditional = self.tracker.within_conditional
            token.in_loop = self.tracker.within_loop
        return line_tokens

    def create_realised_values(self, line_tokens: list[Token]) -> list[RealisedTokens]:
        out: list[RealisedTokens] = []
        for token in line_tokens:
            if is_bool_select(token):
                vals_as_text: list[str] = token.gxparam.get_all_values(nonempty=True) #type: ignore
                vals_as_tlists = [self.create_line_tokens(text) for text in vals_as_text]
                out.append(RealisedTokens(values=vals_as_tlists, original=token))
            else:
                out.append(RealisedTokens(values=[[token]], original=token))
        return out









import yaml
from typing import Optional
from galaxy2janis.paths import USER_DATA_DIR, GALAXY_DATATYPES_YAML

from .JanisDatatype import JanisDatatype


class DatatypeRegister:
    def __init__(self):
        self.dtype_map: dict[str, JanisDatatype] = {}

    def get(self, datatype: str) -> Optional[JanisDatatype]:
        if datatype in self.dtype_map:
            return self.dtype_map[datatype]

    def populate(self) -> None:
        """
        func loads the combined datatype yaml then converts it to dict with format as keys
        provides structue where we can search all the galaxy and janis types given what we see
        in galaxy 'format' attributes.
        """
        path = f'{USER_DATA_DIR}/{GALAXY_DATATYPES_YAML}'
        with open(path, 'r') as fp:
            datatypes = yaml.safe_load(fp)
        for type_data in datatypes['types']:
            janistype = self._init_type(type_data)
            # multiple keys per datatype
            self.dtype_map[type_data['format']] = janistype
            self.dtype_map[type_data['classname']] = janistype 
            if type_data['extensions']:
                for ext in type_data['extensions']:
                    self.dtype_map[ext] = janistype 

    def _init_type(self, dtype: dict[str, str]) -> JanisDatatype:
        return JanisDatatype(
            format=dtype['format'],
            source=dtype['source'],
            classname=dtype['classname'],
            extensions=dtype['extensions'],
            import_path=dtype['import_path']
        )


# SINGLETON
register = DatatypeRegister()







from galaxy2janis.entities.tool import Tool
from galaxy2janis.entities.workflow import Workflow
from galaxy2janis.entities.workflow import WorkflowStep

from .Format import Format


class DefaultFormat(Format):
    """renders output pages in same manner as appears in original workflow"""

    def workflow(self, workflow: Workflow) -> str:
        raise NotImplementedError()
    
    def subworkflow(self, workflow: Workflow) -> str:
        raise NotImplementedError()
    
    def step(self, step: WorkflowStep) -> str:
        raise NotImplementedError()

    def tool(self, tool: Tool) -> str:
        raise NotImplementedError()

    def script(self) -> str:
        raise NotImplementedError()
    
    def inputs(self, workflow: Workflow) -> str:
        raise NotImplementedError()
    
    def config(self, workflow: Workflow) -> str:
        raise NotImplementedError()







from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Optional
from galaxy2janis import expressions
from galaxy2janis.gx.command.components.inputs.InputComponent import InputComponent

from ..gxtool.param.Param import Param 

from .cmdstr.CommandString import CommandString
from .components.CommandComponent import CommandComponent
from .components.inputs.InputComponent import InputComponent
from .components.inputs.Positional import Positional
from .components.inputs.Flag import Flag
from .components.inputs.Option import Option
from .components.outputs.RedirectOutput import RedirectOutput
from .components import factory


class Updater(ABC):
    command: Command
    incoming: Positional | Flag | Option | RedirectOutput
    
    @abstractmethod
    def update(self, command: Command, incoming: Any) -> None:
        """updates the command's store of CommandComponents with the incoming component"""
        ...

    @abstractmethod
    def should_merge(self) -> bool:
        """determines whether to merge the incoming CommandComponent with an existing one, or add new"""
        ...
    
    @abstractmethod
    def merge(self) -> None:
        """updates via merging the incoming CommandComponent with a known CommandComponent"""
        ...
    
    @abstractmethod
    def add(self) -> None:
        """updates via adding a new CommandComponent"""
        ...


class PositionalUpdater(Updater):
    command: Command
    incoming: Positional

    def update(self, command: Command, incoming: Positional) -> None:
        self.command = command
        self.incoming = incoming
        if self.should_merge():
            self.merge()
        else:
            self.add()

    def should_merge(self) -> bool:
        cmd_pos = self.incoming.cmd_pos
        existing_comp = self.command.get_positional(cmd_pos)
        if existing_comp:
            return True
        return False
    
    def merge(self) -> None:
        cmd_pos = self.incoming.cmd_pos
        existing_comp = self.command.get_positional(cmd_pos)
        if existing_comp:
            existing_comp.update(self.incoming)
    
    def add(self) -> None:
        cmd_pos = self.incoming.cmd_pos
        self.command.positionals[cmd_pos] = self.incoming


class FlagUpdater(Updater):
    command: Command
    incoming: Flag

    def update(self, command: Command, incoming: Flag) -> None:
        self.command = command
        self.incoming = incoming
        if self.should_merge():
            self.merge()
        else:
            self.add()

    def should_merge(self) -> bool:
        query_prefix = self.incoming.prefix
        existing_comp = self.command.get_flag(query_prefix)
        if existing_comp:
            return True
        return False
    
    def merge(self) -> None:
        query_prefix = self.incoming.prefix
        existing_comp = self.command.get_flag(query_prefix)
        if existing_comp:
            existing_comp.update(self.incoming)
    
    def add(self) -> None:
        prefix = self.incoming.prefix
        self.command.flags[prefix] = self.incoming


class OptionUpdater(Updater):
    command: Command
    incoming: Option

    def update(self, command: Command, incoming: Option) -> None:
        self.command = command
        self.incoming = incoming
        if self.should_merge():
            self.merge()
        else:
            self.add()

    def should_merge(self) -> bool:
        query_prefix = self.incoming.prefix
        existing_comp = self.command.get_option(query_prefix)
        if existing_comp:
            return True
        return False
    
    def merge(self) -> None:
        query_prefix = self.incoming.prefix
        existing_comp = self.command.get_option(query_prefix)
        if existing_comp:
            existing_comp.update(self.incoming)
    
    def add(self) -> None:
        prefix = self.incoming.prefix
        self.command.options[prefix] = self.incoming


class RedirectOutputUpdater(Updater):
    command: Command
    incoming: RedirectOutput

    def update(self, command: Command, incoming: RedirectOutput) -> None:
        self.command = command
        self.incoming = incoming
        if self.should_merge():
            pass
        else:
            self.add()

    def should_merge(self) -> bool:
        if self.command.redirect:
            return True
        return False
    
    def merge(self) -> None:
        raise RuntimeError("can't update an output")
    
    def add(self) -> None:
        self.command.redirect = self.incoming


class Command:
    def __init__(self, xmlcmdstr: CommandString):
        self.xmlcmdstr = xmlcmdstr
        self.positionals: dict[int, Positional] = {}
        self.flags: dict[str, Flag] = {}
        self.options: dict[str, Option] = {}
        self.redirect: Optional[RedirectOutput] = None

    def gxparam_is_attached(self, gxparam: Param) -> bool:
        components = self.list_inputs()
        components += self.list_outputs()
        for component in components:
            if component.gxparam and component.gxparam.name == gxparam.name:
                return True
        return False

    def list_inputs(self, include_base_cmd: bool=True) -> list[CommandComponent]:
        components: list[CommandComponent] = []
        if include_base_cmd:
            components += self.get_positionals()
        else:
            components += self.get_non_base_positionals()
        components += self.get_flags()
        components += self.get_options()
        return components

    def list_outputs(self) -> list[RedirectOutput]:
        # just returns redirect component if present.
        # other outputs are handled by ToolFactory
        components: list[RedirectOutput] = []
        if self.redirect:
            components.append(self.redirect)
        return components
    
    def get_positional(self, cmd_pos: int) -> Optional[Positional]:
        if cmd_pos in self.positionals:
            return self.positionals[cmd_pos]
        return None

    def get_flag(self, query_prefix: str) -> Optional[Flag]:
        if query_prefix in self.flags:
            return self.flags[query_prefix]
        return None

    def get_option(self, query_prefix: str) -> Optional[Option]:
        if query_prefix in self.options:
            return self.options[query_prefix]
        return None
    
    def get_positionals(self) -> list[Positional]:
        """returns positionals in sorted order"""
        positions_components = list(self.positionals.items())
        positions_components.sort(key=lambda x: x[0])
        return [p[1] for p in positions_components]
    
    def get_flags(self) -> list[Flag]:
        return list(self.flags.values())
    
    def get_options(self) -> list[Option]:
        return list(self.options.values())

    def get_base_positionals(self) -> list[InputComponent]:
        positionals = self.get_positionals()
        out: list[InputComponent] = []
        for p in positionals:
            if self.positional_is_based(p):
                out.append(p)
            else:
                break
        return out

    def positional_is_based(self, p: Positional) -> bool:
        if p.before_opts:
            if not p.gxparam:
                if len(p.values.unique) == 1:
                    if not expressions.is_var(p.values.unique[0]):
                        if not expressions.is_script(p.values.unique[0]):
                            return True
        return False

    def get_non_base_positionals(self) -> list[Positional]:
        base_positionals = self.get_base_positionals()
        all_positionals = self.get_positionals()
        return [p for p in all_positionals if p not in base_positionals]

    def update(self, incoming: CommandComponent) -> None:
        component = self.refine_component(incoming)
        updater = self.select_updater(component)
        updater.update(self, component)

    def refine_component(self, incoming: CommandComponent) -> CommandComponent:
        # migrate incorrect option to flag
        if isinstance(incoming, Option):
            if incoming.prefix in self.flags:
                return factory.flag(incoming.prefix, incoming.gxparam)
        # migrate incorrect flag to option
        # if isinstance(incoming, Flag):
        #     if incoming.prefix in self.options:
        #         return factory.option(prefix=incoming.prefix, gxparam=incoming.gxparam)
        return incoming

    def select_updater(self, incoming: CommandComponent) -> Updater:
        match incoming:
            case Positional():
                return PositionalUpdater()
            case Flag():
                return FlagUpdater()
            case Option():
                return OptionUpdater()
            case RedirectOutput():
                return RedirectOutputUpdater()
            case _:
                raise RuntimeError(f'must pass CommandComponent to Command.update(). received {type(incoming)}')

    def set_cmd_positions(self) -> None:
        options_pos: int = self.get_options_position()
        for flag in self.get_flags():
            flag.cmd_pos = options_pos
        for option in self.get_options():
            option.cmd_pos = options_pos
        
        pos_ptr: int = 1
        for positional in self.get_non_base_positionals():
            if pos_ptr == options_pos:
                pos_ptr += 1
            positional.cmd_pos = pos_ptr
            pos_ptr += 1

    def get_options_position(self) -> int:
        """
        returns cmd_pos for options and flags. 
        base command will always occupy cmd_pos == 0
        """
        i: int = 0
        positionals = self.get_non_base_positionals()
        while i < len(positionals) and positionals[i].before_opts:
            i += 1
        return i + 1

    # string representations
    def __str__(self) -> str:
        return f""" \
##### Command #####

positionals: ------------
{'main value':20}{'optional':>10}
{self._get_components_list_as_str('positional')}

flags: ------------
{'prefix':20}{'optional':>10}
{self._get_components_list_as_str('flag')}

options: ------------
{'prefix':30}{'main value':20}{'optional':>10}
{self._get_components_list_as_str('option')}"""

    def _get_components_list_as_str(self, ctype: str='positional') -> str:
        funcmap = {
            'positional': self.get_positionals,
            'flag': self.get_flags,
            'option': self.get_options,
        }
        outstr: str = ''
        for comp in funcmap[ctype]():
            outstr += comp.__str__() + '\n'
        return outstr
    




from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from typing import Optional

from galaxy2janis import tags
from galaxy2janis.gx.command.components import InputComponent


class InputValueType(Enum):
    RUNTIME         = auto()
    ENV_VAR         = auto()
    STRING          = auto()
    NUMERIC         = auto()
    BOOLEAN         = auto()
    NONE            = auto()

enum_map = {
    #'runtime': InputValueType.RUNTIME,
    'env_var': InputValueType.ENV_VAR,
    'string': InputValueType.STRING,
    'numeric': InputValueType.NUMERIC,
    'boolean': InputValueType.BOOLEAN,
    'none': InputValueType.NONE
}

@dataclass
class InputValue(ABC):
    component: Optional[InputComponent]

    def __post_init__(self):
        self.scatter: bool = False

    @property
    def comptype(self) -> Optional[str]:
        return type(self.component).__name__.lower() 

    @property
    def input_tag(self) -> str:
        """get the str tag for this tool input"""
        if self.component:
            return self.component.tag
        else:
            return 'UNKNOWN'
    
    @property
    @abstractmethod
    def input_value(self) -> str:
        """get the str value for this tool input"""
        ...


@dataclass
class StaticInputValue(InputValue):
    value: str
    _valtypestr: str
    default: bool

    def __post_init__(self):
        self.scatter: bool = False
        self.valtype = enum_map[self._valtypestr]
    
    @property
    def is_none(self) -> bool:
        if self.valtype == InputValueType.NONE:
            return True
        return False
    
    @property
    def input_value(self) -> str:
        if self._should_wrap_value():
            return f'"{self.value}"'
        else:
            return f'{self.value}'

    def _should_wrap_value(self) -> bool:
        if self.valtype == InputValueType.STRING:
            return True
        if self.valtype == InputValueType.ENV_VAR:
            return True
        return False


@dataclass
class ConnectionInputValue(InputValue):
    step_uuid: str
    out_uuid: str
    
    @property
    def input_value(self) -> str:
        step_tag = tags.get(self.step_uuid)
        out_tag = tags.get(self.out_uuid)
        return f'w.{step_tag}.{out_tag}'
    

@dataclass
class WorkflowInputInputValue(InputValue):
    input_uuid: str
    is_runtime: bool

    @property
    def input_value(self) -> str:
        wflow_inp_tag = tags.get(self.input_uuid)
        return f'w.{wflow_inp_tag}'





from datetime import datetime
from typing import Any


from galaxy2janis.runtime.dates import JANIS_DATE_FMT
from galaxy2janis.gx.gxtool.requirements import Requirement


class Wrapper:
    def __init__(self, details: dict[str, Any]):
        self.owner: str = details['owner']
        self.repo: str = details['repo']
        self.revision: str = details['revision']
        self.tool_id: str = details['tool_id']
        self.tool_build: str = details['tool_build']
        self.date_created: datetime = datetime.strptime(details['date_created'], JANIS_DATE_FMT)
        self.requirements: list[Requirement] = details['requirements']
        self.inbuilt: bool = False
        
    @property
    def tool_version(self) -> str:
        return self.tool_build.split('+galaxy')[0]
    
    @property
    def url(self) -> str:
        return f'https://toolshed.g2.bx.psu.edu/repos/{self.owner}/{self.repo}/archive/{self.revision}.tar.gz'

    def to_dict(self) -> dict[str, Any]:
        return {
            'owner': self.owner,
            'repo': self.repo,
            'revision': self.revision,
            'tool_id': self.tool_id,
            'tool_version': self.tool_version,
            'tool_build': self.tool_build,
            'date_created': self.date_created.strftime(JANIS_DATE_FMT),
            'requirements': [req.__dict__ for req in self.requirements],
            'url': self.url,
        }






from dataclasses import dataclass


@dataclass
class Logline:
    """
    represents a single log line within a .log file
    structure = '[%(levelname)s] [%(name)s] %(message)s'
    """
    level: str
    logger: str
    message: str
    contents: str


@dataclass
class Logfile:
    """represents a single .log file"""
    subtype: str    
    path: str
    lines: list[Logline]

    @property
    def messages(self) -> list[str]:
        messages = [line.message for line in self.lines]
        return list(set(messages))





from galaxy2janis.logs import logging
import tempfile
from typing import Optional

from galaxy2janis.gx.gxtool.requirements import Requirement, CondaRequirement, ContainerRequirement
from galaxy2janis.paths import USER_DATA_DIR, CONTAINER_CACHE

from .ContainerCache import ContainerCache
from .Container import Container

from .fetching.Fetcher import Fetcher
from .fetching.ContainerReqFetcher import ContainerReqFetcher
#from .fetching.GA4GHFetcher import GA4GHFetcher
from .fetching.QuayIOFetcher import QuayIOFetcher
from .fetching.presets import get_images_preset

from .selection.selection import select_best_container_match




DISABLE_CACHE = False

def fetch_container(requirement: Requirement) -> Optional[str]:
    containers: list[Container] = []
    if not containers:
        containers = _fetch_cache(requirement)
    if not containers:
        containers = _fetch_presets(requirement)
    if not containers:
        containers = _fetch_online(requirement)
    if not containers:
        logging.no_container()
        return None
    else:
        container = select_best_container_match(containers, requirement)
        return container.url
    
def _fetch_cache(requirement: Requirement) -> list[Container]:
    cache = _load_cache()
    return cache.get(requirement.name)

def _load_cache() -> ContainerCache:
    if DISABLE_CACHE:
        temp = tempfile.TemporaryFile()
        cache_path = f'{tempfile.gettempdir()}/{temp.name}'
    else:
        cache_path = f'{USER_DATA_DIR}/{CONTAINER_CACHE}'
    return ContainerCache(cache_path)

def _fetch_presets(requirement: Requirement) -> list[Container]:
    return get_images_preset(requirement)

def _fetch_online(requirement: Requirement) -> list[Container]:
    strategy = _select_strategy(requirement)
    containers = strategy.fetch(requirement)
    if containers:
        _update_cache(containers)
    return containers
    
def _select_strategy(requirement: Requirement) -> Fetcher:
    match requirement:
        case CondaRequirement():
            return QuayIOFetcher()
        #case CondaRequirement():
        #    return GA4GHFetcher()
        case ContainerRequirement():
            return ContainerReqFetcher()
        case _:
            raise RuntimeError()
    
def _update_cache(containers: list[Container]) -> None:
    cache = _load_cache()
    for container in containers: 
        cache.add(container)









DEFAULT_TOOL_OUTDIR = 'parsed/tools'
DEFAULT_WORKFLOW_OUTDIR = 'parsed/workflows'

COMMON_SUBFOLDERS: list[str] = [
    'logs',
]
WORKFLOW_SUBFOLDERS: list[str] = [
    'subworkflows',
    'tools',
    'tools/wrappers',
    'tools/scripts',
    'tools/untranslated'
]

from pathlib import Path
home = str(Path.home())

USER_DATA_DIR = f'{home}/.galaxy2janis'
LOGGING_CONFIG = 'logging_config.yaml'
GALAXY_CONFIG = 'galaxy_config.yaml'
GALAXY_DATATYPES_YAML = 'galaxy_janis_types.yaml'
CONTAINER_CACHE = 'container_url_cache.json'
WRAPPER_CACHE = 'wrappers.json'
DOWNLOADED_WRAPPERS_DIR = 'wrappers'




from typing import Optional

_outdir: str
_tool_subdir: Optional[str] = None


def set_outdir(value: str) -> None:
    global _outdir
    _outdir = value

def set_tool_subdir(value: Optional[str]) -> None:
    global _tool_subdir
    _tool_subdir = value


# runtime files
def outdir() -> str:
    return _outdir

def common_subfolders() -> list[str]:
    return [f'{_outdir}/{sf}' for sf in COMMON_SUBFOLDERS]

def workflow_subfolders() -> list[str]:
    return [f'{_outdir}/{sf}' for sf in WORKFLOW_SUBFOLDERS]

def janis_log() -> str:
    return f'{_outdir}/logs/janis.log'

def message_log() -> str:
    return f'{_outdir}/logs/messages.log'


# workflow files
def tool(tool_id: str) -> str:
    if _tool_subdir:
        return f'{_outdir}/{_tool_subdir}/{tool_id}.py'
    else:
        return f'{_outdir}/{tool_id}.py'

def workflow() -> str:
    return f'{_outdir}/workflow.py'

def inputs(file_format: str='yaml') -> str:
    return f'{_outdir}/inputs.{file_format}'

def config(file_format: str='yaml') -> str:
    return f'{_outdir}/config.{file_format}'

def subworkflow(tool_id: str) -> str:
    return f'{_outdir}/subworkflows/{tool_id}.py'

def untranslated(tool_id: str) -> str: 
    return f'{_outdir}/tools/untranslated/{tool_id}.txt'

def configfile(tool_id: str, configfile_name: str) -> str:
    return f'{_outdir}/tools/scripts/{tool_id}_{configfile_name}'

def script(filename: str) -> str:
    return f'{_outdir}/tools/scripts/{filename}'

def wrapper(tool_id: str, revision: str) -> str:
    return f'{_outdir}/tools/wrappers/{tool_id}-{revision}'






import textwrap

from typing import Tuple
from datetime import datetime

from galaxy2janis.entities.tool.Tool import Tool
from galaxy2janis.runtime.dates import JANIS_DATE_FMT

from ..tool.ToolInputSectionText import ToolInputSectionText
from ..tool.ToolOutputSectionText import ToolOutputSectionText
from ..TextRender import TextRender

from .. import ordering
from .. import formatting


def note_snippet(tool: Tool) -> str:
    tool_name = tool.metadata.id
    tool_version = tool.metadata.version
    return f"""\
# NOTE
# This is an automated translation of the '{tool_name}' version '{tool_version}' tool from a Galaxy XML tool wrapper.  
# Translation was performed by the gxtool2janis program (in-development)
"""

def syspath_snippet() -> str:
    return """\
import sys
sys.path.append('/home/grace/work/pp/gxtool2janis')
"""

def metadata_snippet(tool: Tool) -> str:
    return f"""\
metadata = ToolMetadata(
    short_documentation="{tool.metadata.description}",
    keywords=[],
    contributors={get_contributors(tool)},
    dateCreated="{datetime.today().strftime(JANIS_DATE_FMT)}",
    dateUpdated="{datetime.today().strftime(JANIS_DATE_FMT)}",
    version="{tool.metadata.version}",
    doi="{tool.metadata.get_doi_citation()}",
    citation="{tool.metadata.get_main_citation()}",
    documentationUrl=None,
    documentation=\"\"\"{str(tool.metadata.help)}\"\"\"
)
"""

def get_contributors(tool: Tool) -> list[str]:
    contributors: list[str] = ['gxtool2janis']
    if tool.metadata.owner:
        contributors += [f'Wrapper owner: galaxy toolshed user {tool.metadata.owner}']
    if tool.metadata.creator:
        contributors += [f'Wrapper creator: {tool.metadata.creator}']
    return contributors

def builder_snippet(tool: Tool) -> str:
    container = f'"{tool.container}"' if tool.container else None
    return f"""\
{tool.tag} = CommandToolBuilder(
    tool="{tool.tag}",
    version="{tool.metadata.version}",
    metadata=metadata,
    container={container},
    base_command={tool.base_command},
    inputs=inputs,
    outputs=outputs
)
"""

def translate_snippet(tool: Tool) -> str:
    tool_tag = tool.tag
    return textwrap.dedent(f"""\
    if __name__ == "__main__":
        {tool_tag}().translate(
            "wdl", to_console=True
        )\n
    """
    )


core_imports = [
    ("janis_core", "CommandToolBuilder"),
    ("janis_core", "ToolMetadata"),
    ("janis_core", "ToolInput"),
    ("janis_core", "ToolOutput"),
]


class ToolText(TextRender):
    def __init__(self, entity: Tool):
        super().__init__()
        self.entity = entity

    @property
    def imports(self) -> list[Tuple[str, str]]:
        inputs = self.entity.inputs
        outputs = self.entity.outputs
        imports: list[Tuple[str, str]] = []
        imports += core_imports
        imports += ToolInputSectionText(inputs).imports
        imports += ToolOutputSectionText(outputs).imports
        imports = list(set(imports))
        return ordering.order_imports(imports)

    def render(self) -> str:
        out_str: str = ''
        out_str += f'{note_snippet(self.entity)}\n'
        # messages here?
        out_str += f'{syspath_snippet()}\n'
        out_str += f'{formatting.format_imports(self.imports)}\n'
        out_str += f'\n{metadata_snippet(self.entity)}\n'
        out_str += f'{ToolInputSectionText(self.entity.inputs).render()}\n'
        out_str += f'{ToolOutputSectionText(self.entity.outputs).render()}\n'
        out_str += f'{builder_snippet(self.entity)}\n'
        out_str += f'\n{translate_snippet(self.entity)}\n'
        return out_str





from typing import Optional, Tuple
from dataclasses import dataclass

from galaxy2janis.gx.command.components import Flag
from galaxy2janis.gx.command.components import InputComponent
from galaxy2janis.gx.command.components import Positional
from galaxy2janis.gx.command.components import Option

from ..TextRender import TextRender
from .. import formatting
from .. import ordering


@dataclass
class ToolInputText(TextRender):
    def __init__(self, entity: InputComponent):
        super().__init__()
        self.entity = entity

    @property
    def imports(self) -> list[Tuple[str, str]]:
        jtype = self.entity.datatype
        imports: list[Tuple[str, str]] = []
        imports.append((jtype.import_path, jtype.classname))

        # TODO opportunity for decorator
        if self.entity.array:
            imports.append(('janis_core', 'Array'))
        
        # TODO opportunity for decorator
        imports = list(set(imports))
        return ordering.order_imports(imports)

    def render(self) -> str:
        e = self.entity
        prefix = self.format_component_prefix()
        kv_space = self.format_separation()
        doc = formatting.format_docstring(e)
        datatype_str = formatting.format_typestr(e)
        out_str: str = ''
        out_str += '\tToolInput(\n'
        out_str += f"\t\t'{e.tag}',\n"
        out_str += f"\t\t{datatype_str},\n" 
        out_str += f"\t\tprefix='{prefix}',\n" if prefix else ''
        out_str += f"\t\tseparate_value_from_prefix={kv_space},\n" if kv_space == False else ''
        out_str += f"\t\tposition={e.cmd_pos},\n"
        out_str += f"\t\tdefault={formatting.get_wrapped_default(e)},\n"
        out_str += f'\t\tdoc="{doc}",\n' if doc else ''
        out_str += '\t)'
        return out_str

    def format_component_prefix(self) -> Optional[str]:
        e = self.entity
        match e:
            case Positional():
                return None
            case Flag():
                return e.prefix
            case Option():
                return e.prefix if e.delim == ' ' else e.prefix + e.delim
            case _:
                raise RuntimeError
    
    def format_separation(self) -> bool:
        if isinstance(self.entity, Option):
            if self.entity.delim == ' ':
                return True
            else:
                return False
        else:
            return True






from dataclasses import dataclass
from typing import Any, Optional

from galaxy2janis.gx.wrappers import Wrapper


@dataclass
class StepMetadata:
    wrapper: Wrapper
    step_id: int
    step_name: str
    tool_state: dict[str, Any]
    workflow_outputs: list[dict[str, Any]]
    _label: Optional[str] = None

    @property
    def label(self) -> str:
        if self._label:
            return self._label
        return self.step_name






from copy import deepcopy
from typing import Iterable

from ..tokens import Token
from ..epath.ExecutionPath import ExecutionPath
from .RealisedTokenValues import RealisedTokens


class DynamicCommandStatement:
    """
    represents a command line statement. 
    
    for example:
        ln -sf '${file_input}' ${sample_name} &&
        abricate ${sample_name} > '$report'
    
    in the above, there are two statements: the symlink, then the abricate command.
    these are seperated by &&.

    one of these statements will be the tool actually being executed, the other is pre/post
    processing. 
    
    each individual command line statement is represented as a DynamicCommandStatement.
    each statement can have multiple realised values, in the case of galaxy select or bool 
    params being present in that statement statement.
    """

    def __init__(self, cmdline: str, realised_tokens: list[RealisedTokens]):
        self.cmdline = cmdline
        self.realised_tokens = realised_tokens

    def get_tokens(self) -> list[Token]:
        return [rt.get_original_token() for rt in self.realised_tokens]

    def get_execution_paths(self) -> Iterable[ExecutionPath]:
        """
        galaxy select or bool params may have multiple possible realised values 
        in a command string. for each of these values, we create an ExecutionPath
        which represents a possible realised command line, with a particular value 
        slotted in for that select or bool param. 
                
        Each other token becomes the default value for that position 
        (ie abricate --$input1 will always appear as Token(abricate), Token(--$input1) 
        as there are no alternatives). 
        
        For sitations where the other token is also a bool / select, we just provide 
        any realised value of that bool / select as a token. 

        each of these ExecutionPaths are fed into the program later when attempting
        to understand the tool Command()

        """ 
        # get all tokens as defaults
        default_tokens = [rtvs.get_default_token() for rtvs in self.realised_tokens]
        for i in range(len(self.realised_tokens)):
            tokens_copy = deepcopy(default_tokens)
            if len(self.realised_tokens[i].tlists) > 1:
                # get the additional different values the current position can take and 
                # hot swap these in to the default tokens to produce 'realised' tokens
                for tlist in self.realised_tokens[i].tlists[1:]:
                    realised_tokens = tokens_copy[:i] + tlist + tokens_copy[i + 1:]
                    yield ExecutionPath(realised_tokens)

        # yield the defaults as final ExecutionPath
        yield ExecutionPath(default_tokens)
    
    def get_galaxy_reference_count(self) -> int:
        default_tokens = [rtvs.get_default_token() for rtvs in self.realised_tokens]
        return sum([token.gxparam is not None for token in default_tokens])
    
    def get_first_word(self) -> str:
        return self.realised_tokens[0].tlists[0][0].text # pure cancer but 






import json
from typing import Any

from galaxy2janis.entities.workflow import WorkflowInput
from galaxy2janis.entities.workflow import Workflow

from galaxy2janis import datatypes


def ingest_workflow_inputs(janis: Workflow, galaxy: dict[str, Any]) -> None:
    for step in galaxy['steps'].values():
        if step['type'] in ['data_input', 'data_collection_input']:
            workflow_input = parse_input_step(step)
            step['janis_uuid'] = workflow_input.uuid # crucial!
            janis.add_input(workflow_input)

def parse_input_step(step: dict[str, Any]) -> WorkflowInput:
    return WorkflowInput(
        _name=format_input_step_name(step),
        array=format_input_step_array(step),
        is_runtime=False,
        datatype=datatypes.get(step, 'GalaxyInputStep'),
        optional=format_input_step_optionality(step)
    )

def format_input_step_name(step: dict[str, Any]) -> str:
    if step['label'] and step['label'] != '':
        return step['label']
    elif step['inputs']:
        return step['inputs'][0]['name']
    else:
        raise RuntimeError()
    
def format_input_step_array(step: dict[str, Any]) -> bool:
    if step['type'] and step['type'] == 'data_collection_input':
        return True
    return False

def format_input_step_optionality(step: dict[str, Any]) -> bool:
    tool_state = json.loads(step['tool_state'])
    if 'optional' in tool_state:
        return bool(tool_state['optional'])
    return False





from typing import Optional

from ..step.inputs import InputValue, WorkflowInputInputValue
from ..step.inputs import ConnectionInputValue



class StepInputRegister:
    def __init__(self):
        self.inputs: list[InputValue] = []

    @property
    def all(self) -> list[InputValue]:
        return self.inputs
    
    @property
    def connections(self) -> list[ConnectionInputValue]:
        return [x for x in self.inputs if isinstance(x, ConnectionInputValue)]
    
    @property
    def workflow_inputs(self) -> list[WorkflowInputInputValue]:
        return [x for x in self.inputs if isinstance(x, WorkflowInputInputValue)]
    
    @property
    def linked(self) -> list[InputValue]:
        return [x for x in self.inputs if x.component]
    
    @property
    def unlinked(self) -> list[InputValue]:
        return [x for x in self.inputs if not x.component]

    def add(self, invalue: InputValue) -> None:
        self.inputs.append(invalue)
    
    def get(self, query_uuid: str) -> Optional[InputValue]:
        for invalue in self.inputs:
            if invalue.component and invalue.component.uuid == query_uuid:
                return invalue





from typing import Any
from abc import ABC, abstractmethod
from . import rules as rules


class FormattingStrategy(ABC):

    @abstractmethod
    def format(self, starting_text: str, entity: Any) -> str:
        ...

# default
class GenericFormattingStrategy(FormattingStrategy):
    def format(self, starting_text: str, entity: Any) -> str:
        tag = starting_text
        tag = rules.numeric(tag, entity)
        tag = rules.numeric_start(tag, entity)
        tag = rules.non_alphanumeric(tag, entity)
        tag = rules.short_tag(tag, entity)
        tag = rules.camelify(tag)
        #tag = rules.encode(tag)
        tag = rules.replace_keywords(tag, entity)
        return tag

class WorkflowInputFormattingStrategy(FormattingStrategy):
    def format(self, starting_text: str, entity: Any) -> str:
        tag = starting_text
        if not entity.is_runtime:
            tag = rules.numeric(tag, entity)
            tag = rules.numeric_start(tag, entity)
            tag = rules.non_alphanumeric(tag, entity)
            tag = rules.short_tag(tag, entity)
            #tag = rules.encode(tag)
            tag = rules.camelify(tag)
            tag = rules.replace_keywords(tag, entity)
        return tag

# capitalisation is allowed
class ToolNameStrategy(FormattingStrategy):
    def format(self, starting_text: str, entity: Any) -> str:
        tag = starting_text
        tag = rules.non_alphanumeric(tag, entity)
        tag = rules.numeric(tag, entity)
        tag = rules.numeric_start(tag, entity)
        tag = rules.short_tag(tag, entity)
        #tag = rules.encode(tag)
        tag = rules.camelify(tag)
        tag = rules.replace_keywords(tag, entity)
        return tag


STRATEGIES = {
    'Workflow': GenericFormattingStrategy(),
    'WorkflowInput': WorkflowInputFormattingStrategy(),
    'WorkflowStep': GenericFormattingStrategy(),
    'Tool': ToolNameStrategy(),
    'Positional': GenericFormattingStrategy(),
    'Flag': GenericFormattingStrategy(),
    'Option': GenericFormattingStrategy(),
    'RedirectOutput': GenericFormattingStrategy(),
    'WildcardOutput': GenericFormattingStrategy(),
    'InputOutput': GenericFormattingStrategy(),
}


def format_tag(starting_text: str, entity: Any) -> str:
    strategy = STRATEGIES[entity.__class__.__name__]
    return strategy.format(starting_text, entity)










from __future__ import annotations
from typing import TYPE_CHECKING, Tuple

if TYPE_CHECKING:
    from galaxy2janis.entities.workflow import Workflow

from datetime import datetime
from galaxy2janis.runtime.dates import JANIS_DATE_FMT

from .. import formatting
from .. import ordering

from ..TextRender import TextRender
from .WorkflowInputText import WorkflowInputText
from .WorkflowOutputText import WorkflowOutputText
from .StepText import StepText


### HELPER METHODS ### 

def note_snippet(workflow: Workflow) -> str:
    return f"""# NOTE
# This is an automated translation of the '{workflow.metadata.name}' version '{workflow.metadata.version}' workflow. 
# Translation was performed by the gxtool2janis program (in-development)
"""

def inputs_yaml_snippet() -> str:
    return f"""
\"\"\"
NOTE
Please provide values for these workflow inputs 
in inputs.yaml
\"\"\"
"""

def syspath_snippet() -> str: # ???
    return """import sys
sys.path.append('/home/grace/work/pp/gxtool2janis')
"""

def metadata_snippet(workflow: Workflow) -> str:
    contributors = ['gxtool2janis']
    return f"""metadata = WorkflowMetadata(
    short_documentation="{workflow.metadata.annotation}",
    contributors={contributors},
    keywords={workflow.metadata.tags},
    dateCreated="{datetime.today().strftime(JANIS_DATE_FMT)}",
    dateUpdated="{datetime.today().strftime(JANIS_DATE_FMT)}",
    version={workflow.metadata.version}
)
"""

def builder_snippet(workflow: Workflow) -> str:
    out_str: str = ''
    out_str += 'w = WorkflowBuilder(\n'
    out_str += f'\t"{workflow.tag}",\n'
    out_str += f'\tversion="{workflow.metadata.version}",\n'
    out_str += f'\tdoc="{workflow.metadata.annotation}"\n'
    out_str += ')\n'
    return out_str

def translate_snippet() -> str:
    return f"""if __name__ == "__main__":
    w.translate("cwl", **args)
"""

core_imports = [
    ("janis_core", "WorkflowMetadata"),
    ("janis_core", "WorkflowBuilder"),
]

def label(text: str) -> str: 
    title = f'# {text}'
    border = f'# {"-" * (len(title) - 2)}'
    return f"""\
{border}
{title}
{border}"""

### MAIN CLASS ### 

class WorkflowText(TextRender):
    def __init__(
        self, 
        entity: Workflow, 
        render_note: bool=True,
        render_syspath: bool=True,
        render_translate: bool=True
    ):
        super().__init__()
        self.entity = entity
        self.render_note = render_note
        self.render_syspath = render_syspath
        self.render_translate = render_translate

    @property
    def imports(self) -> list[Tuple[str, str]]:
        imports: list[Tuple[str, str]] = []
        imports += core_imports
        # imports for each workflow input (eg datatypes)
        for winp in self.entity.inputs:
            imports += WorkflowInputText(winp).imports
        # imports for each translated tool
        for step in self.entity.steps:
            tool_id = step.metadata.wrapper.tool_id
            relative_path = f'tools.{tool_id}'
            tool_tag = step.tool.tag
            imports.append((relative_path, tool_tag))
        # imports used in steps
        for step in self.entity.steps:
            imports += StepText(step, self.entity).imports
        # imports for each output (mostly datatypes)
        for wout in self.entity.outputs:
            imports += WorkflowOutputText(wout).imports
        imports = list(set(imports))
        return ordering.order_imports(imports)

    def render(self) -> str:
        out_str: str = ''
        if self.render_note:
            out_str += f'{note_snippet(self.entity)}\n'
        # messages here?
        if self.render_syspath:
            out_str += f'{syspath_snippet()}\n'
        out_str += f'{formatting.format_imports(self.imports)}\n'
        out_str += f'{metadata_snippet(self.entity)}\n'
        out_str += f'{builder_snippet(self.entity)}\n'
        
        # inputs
        out_str += f"{label('INPUTS')}\n"
        out_str += f"{inputs_yaml_snippet()}\n"
        for winp in self.entity.inputs:
            #if not winp.is_runtime:
            out_str += f'{WorkflowInputText(winp).render()}\n'
        out_str += '\n'
        
        # steps (includes tool steps and subworkflow steps)
        for i, step in enumerate(self.entity.steps):
            out_str += f"{label(f'STEP{i+1}: {step.metadata.label}')}\n"
            out_str += '\n'
            out_str += f'{StepText(step, self.entity).render()}\n'
        
        # outputs
        out_str += f"{label('OUTPUTS')}\n\n"
        for wout in self.entity.outputs:
            out_str += f'{WorkflowOutputText(wout).render()}\n'

        if self.render_translate:
            out_str += f'{translate_snippet()}\n'

        return out_str
    





from galaxy2janis.logs import logging
from galaxy2janis.gx.gxtool import XMLToolDefinition
from galaxy2janis.gx.command import Command
from typing import Optional

# this module imports
from .Tool import Tool
from .outputs import extract_outputs


class ToolFactory:
    def __init__(self, xmltool: XMLToolDefinition, command: Command, container: Optional[str]) -> None:
        self.xmltool = xmltool
        self.command = command
        self.container = container

    def create(self) -> Tool:
        tool = Tool(
            metadata=self.xmltool.metadata,
            container=self.container,
            base_command=self.get_base_command(),
            gxparam_register=self.xmltool.inputs,
            configfiles=self.xmltool.configfiles
        )
        self.supply_inputs(tool)
        self.supply_outputs(tool)
        return tool

    def supply_inputs(self, tool: Tool) -> None:
        self.command.set_cmd_positions()
        inputs = self.command.list_inputs(include_base_cmd=False)
        if not inputs:
            logging.no_inputs()
        for inp in inputs:
            tool.add_input(inp)

    def supply_outputs(self, tool: Tool) -> None:
        outputs = extract_outputs(self.xmltool, self.command)
        if not outputs:
            logging.no_outputs()
        for out in outputs:
            tool.add_output(out)

    def get_base_command(self) -> list[str]:
        positionals = self.command.get_base_positionals()
        if not positionals:
            logging.no_base_cmd()
        return [p.default_value for p in positionals]
    



from __future__ import annotations
from typing import Any, Optional

from galaxy2janis.gx.gxtool.param import BoolParam
from galaxy2janis.gx.gxtool.param import Param

from .InputComponent import InputComponent


class Flag(InputComponent):
    def __init__(self, prefix: str) -> None:
        super().__init__()
        self.prefix = prefix

    @property
    def name(self) -> str:
        return self.prefix.strip('--')

    @property
    def default_value(self) -> bool:
        if isinstance(self.gxparam, BoolParam):
            if self.gxparam.checked and self.gxparam.truevalue == self.prefix:
                return True
            elif self.gxparam.checked and self.gxparam.truevalue == "":
                return False
            elif not self.gxparam.checked and self.gxparam.falsevalue == self.prefix:
                return True
            elif not self.gxparam.checked and self.gxparam.falsevalue == "":
                return False
        return False
    
    @property
    def optional(self) -> bool:
        return True

    @property
    def array(self) -> bool:
        return False

    @property
    def docstring(self) -> Optional[str]:
        if self.gxparam:
            return self.gxparam.docstring
        return None

    def update(self, incoming: Any):
        assert(isinstance(incoming, Flag))
        # gxparam transfer
        if not self.gxparam and incoming.gxparam:
            self.gxparam: Optional[Param] = incoming.gxparam
        
    def __str__(self) -> str:
        return f'{str(self.default_value):20}{str(self.optional):>10}'




import os
import shutil
import tempfile
from typing import Any

from sqlalchemy.orm.scoping import scoped_session

from galaxy import (
    di,
    model,
    objectstore,
    quota,
)
from galaxy.tools.parameters.basic import (
    IntegerToolParameter,
)
from galaxy.auth import AuthManager
from galaxy.datatypes.registry import Registry
from galaxy.jobs.manager import NoopManager
from galaxy.jobs import SimpleComputeEnvironment
from galaxy.managers.users import UserManager
from galaxy.model import mapping, tags
from galaxy.model.base import SharedModelMapping
from galaxy.model.mapping import GalaxyModelMapping
from galaxy.security import idencoding
from galaxy.structured_app import BasicApp, MinimalManagerApp, StructuredApp
from galaxy.tool_util.deps.containers import NullContainerFinder
from galaxy.util import StructuredExecutionTimer
from galaxy.util import galaxy_directory
from galaxy.util.bunch import Bunch
from galaxy.util.dbkeys import GenomeBuilds
from galaxy.web_stack import ApplicationStack
from galaxy.util import XML
from galaxy.tools.parameters import params_from_strings # for MockTool
from galaxy.tool_util.parser.output_objects import ToolOutput
from galaxy.tool_util.biotools import BiotoolsMetadataSource
from galaxy.tools.data import ToolDataTableManager
#from galaxy.datatypes.registry import example_datatype_registry_for_sample
# terrible stuff from galaxy. not planning on a refactor.

import yaml
from galaxy2janis import settings




class MockApp(di.Container):
    def __init__(self, config=None, **kwargs):
        super().__init__()
        self[BasicApp] = self
        self[MinimalManagerApp] = self
        self[StructuredApp] = self
        self.config = config or MockAppConfig(**kwargs)
        self.security = self.config.security
        self[idencoding.IdEncodingHelper] = self.security
        self.name = kwargs.get('name', 'galaxy')
        self.object_store = objectstore.build_object_store_from_config(self.config)
        self.model = mapping.init("/tmp", self.config.database_connection, create_tables=True, object_store=self.object_store)
        self[SharedModelMapping] = self.model
        self[GalaxyModelMapping] = self.model
        self[scoped_session] = self.model.context
        self.security_agent = self.model.security_agent
        self.visualizations_registry = MockVisualizationsRegistry()
        self.tag_handler = tags.GalaxyTagHandler(self.model.context)
        self[tags.GalaxyTagHandler] = self.tag_handler
        self.quota_agent = quota.DatabaseQuotaAgent(self.model)
        self.init_datatypes()
        # self.job_config = Bunch(
        #     dynamic_params=None,
        #     destinations={},
        #     use_messaging=False,
        #     assign_handler=lambda *args, **kwargs: None
        # )
        self.tool_data_tables = self.grace_init_data_tables()
        self.dataset_collections_service = None
        self.container_finder = NullContainerFinder()
        self._toolbox_lock = MockLock()
        self.tool_shed_registry = Bunch(tool_sheds={})
        self.genome_builds = GenomeBuilds(self)
        self.job_manager = NoopManager()
        self.application_stack = ApplicationStack()
        self.auth_manager = AuthManager(self.config)
        self.user_manager = UserManager(self)
        self.execution_timer_factory = Bunch(get_timer=StructuredExecutionTimer)
        self.is_job_handler = False
        self.biotools_metadata_source = BiotoolsMetadataSource()
        # new param not native to the aApp class
        self.dataset_counter: int = 1

        def url_for(*args, **kwds):
            return "/mock/url"
        self.url_for = url_for

    def grace_init_data_tables(self) -> ToolDataTableManager | dict[str, str]:
        # Initialize tool data tables using the config defined by self.config.tool_data_table_config_path.
        if self.config.tool_data_path and self.config.tool_data_table_config_path:
            return ToolDataTableManager(
                tool_data_path=self.config.tool_data_path,
                config_filename=self.config.tool_data_table_config_path,
                other_config_dict=self.config
            )
        return {}

    def init_datatypes(self):
        # config = Bunch(sniff_compressed_dynamic_datatypes_default=True)
        # datatypes_registry = registry.Registry(config=config)
        # datatypes_registry.load_datatypes(root_dir='galaxy', config='config/datatypes_conf.xml.sample')
        # datatypes_registry = example_datatype_registry_for_sample()
        # model.set_datatypes_registry(datatypes_registry)
        # self.datatypes_registry = datatypes_registry
        galaxy_dir = galaxy_directory()
        sample_conf = os.path.join(galaxy_dir, "site-packages", "galaxy", "config", "sample", "datatypes_conf.xml.sample")
        config = Bunch(sniff_compressed_dynamic_datatypes_default=True)
        datatypes_registry = Registry(config)
        datatypes_registry.load_datatypes(root_dir=galaxy_dir, config=sample_conf)
        model.set_datatypes_registry(datatypes_registry)
        self.datatypes_registry = datatypes_registry

    def wait_for_toolbox_reload(self, toolbox):
        # TODO: If the tpm test case passes, does the operation really
        # need to wait.
        return True



from galaxy2janis.paths import USER_DATA_DIR, GALAXY_CONFIG

def grace_get_config() -> dict[str, Any]:
    config_path = f'{USER_DATA_DIR}/{GALAXY_CONFIG}'
    with open(config_path, "r") as fp:
        return yaml.safe_load(fp)

class MockAppConfig(Bunch):

    class MockSchema(Bunch):
        pass

    def __init__(self, root=None, **kwargs):
        Bunch.__init__(self, **kwargs)
        if not root:
            root = tempfile.mkdtemp()
            self._remove_root = True
        else:
            self._remove_root = False   
        self.schema = self.MockSchema()
        self.security = idencoding.IdEncodingHelper(id_secret='6e46ed6483a833c100e68cc3f1d0dd76')
        self.database_connection = kwargs.get('database_connection', "sqlite:///:memory:")
        self.use_remote_user = kwargs.get('use_remote_user', False)
        self.enable_celery_tasks = False
        self.data_dir = os.path.join(root, 'database')
        self.file_path = os.path.join(self.data_dir, 'files')
        self.jobs_directory = os.path.join(self.data_dir, 'jobs_directory')
        self.new_file_path = os.path.join(self.data_dir, 'tmp')

        # GRACE TOOL DATA
        self.tool_data_path = None
        self.tool_data_table_config_path = None
        self.tool_dependency_dir = None
        self.set_tool_data_attrs()

        # GRACE LOGGING
        self.logging = grace_get_config()

        self.metadata_strategy = 'legacy'

        self.object_store_config_file = ''
        self.object_store = 'disk'
        self.object_store_check_old_style = False
        self.object_store_cache_path = '/tmp/cache'
        self.umask = os.umask(0o77)
        self.gid = os.getgid()

        self.user_activation_on = False
        self.new_user_dataset_access_role_default_private = False

        self.expose_dataset_path = True
        self.allow_user_dataset_purge = True
        self.enable_old_display_applications = True
        self.redact_username_in_logs = False
        self.auth_config_file = "config/auth_conf.xml.sample"
        self.error_email_to = "admin@email.to"
        self.password_expiration_period = 0

        self.umask = 0o77
        self.flush_per_n_datasets = 0

        # Compliance related config
        self.redact_email_in_job_name = False

        # Follow two required by GenomeBuilds
        self.len_file_path = os.path.join('tool-data', 'shared', 'ucsc', 'chrom')
        self.builds_file_path = os.path.join('tool-data', 'shared', 'ucsc', 'builds.txt.sample')

        self.shed_tool_config_file = "config/shed_tool_conf.xml"
        self.shed_tool_config_file_set = False
        self.enable_beta_edam_toolbox = False
        self.preserve_python_environment = "always"
        self.enable_beta_gdpr = False

        self.version_major = "19.09"

        # set by MockDir
        self.root = root
        self.enable_tool_document_cache = False
        self.tool_cache_data_dir = os.path.join(root, 'tool_cache')
        self.delay_tool_initialization = True
        self.external_chown_script = None

        self.default_panel_view = "default"
        self.panel_views_dir = ''
        self.panel_views = {}
        self.edam_panel_views = ''

        self.config_file = None

    def set_tool_data_attrs(self) -> None:
        tool_data = f'{settings.tool.xml_dir()}/tool-data'
        tool_data_table_conf = f'{settings.tool.xml_dir()}/tool_data_table_conf.xml.sample'
        if os.path.exists(tool_data):
            self.tool_data_path = tool_data
        if os.path.exists(tool_data_table_conf):
            self.tool_data_table_config_path = tool_data_table_conf

    @property
    def config_dict(self):
        return self.dict()

    def __getattr__(self, name):
        # Handle the automatic [option]_set options: for tests, assume none are set
        if name == 'is_set':
            return lambda x: False
        # Handle the automatic config file _set options
        if name.endswith('_file_set'):
            return False
        raise AttributeError(name)

    def __del__(self):
        if self._remove_root:
            shutil.rmtree(self.root)



class MockVisualizationsRegistry:
    BUILT_IN_VISUALIZATIONS = ['trackster']

    def get_visualizations(self, trans, target):
        return []



class MockLock:
    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        pass



class MockObjectStore:
    def __init__(self):
        self.created_datasets = []
        self.first_create = True
        self.object_store_id = "mycoolid"

    def exists(self, *args, **kwargs):
        return True

    def create(self, dataset):
        self.created_datasets.append(dataset)
        if self.first_create:
            self.first_create = False
            assert dataset.object_store_id is None
            dataset.object_store_id = self.object_store_id
        else:
            assert dataset.object_store_id == self.object_store_id


class ComputeEnvironment(SimpleComputeEnvironment):
    def __init__(
        self,
        new_file_path,
        working_directory
    ):
        self._tool_dir = ''
        self._new_file_path = new_file_path
        self._working_directory = working_directory
        self._unstructured_path_rewrites = {}

    def input_path_rewrite(self, dataset):
        return dataset.file_name.rsplit('/', 1)[-1]

    def output_path_rewrite(self, dataset):
        return dataset.file_name.rsplit('/', 1)[-1]

    def output_paths(self):
        return self._output_paths

    def working_directory(self):
        return self._working_directory

    def home_directory(self):
        return self._working_directory

    def tmp_directory(self):
        return self._working_directory

    def new_file_path(self):
        return self._new_file_path

    def unstructured_path_rewrite(self, path):
        for key, val in self._unstructured_path_rewrites.items():
            if path.startswith(key):
                return path.replace(key, val)
        return None

    def tool_directory(self):
        return self._tool_dir

    def galaxy_url(self):
        return 'http://localhost:9090/'


# not used?
class MockTool:
    def __init__(self, app: MockApp):
        self.profile = 16.01
        self.python_template_version = '2.7'
        self.app = app
        self.hooks_called = []
        self.environment_variables = []
        self._config_files = []
        self._command_line = "bwa --thresh=$thresh --in1=$input1 --in2=$input2 --out=$output1"
        self._params = {"thresh": self.test_thresh_param()}
        self.options = Bunch(sanitize=False)
        self.check_values = True

    def test_thresh_param(self):
        elem = XML('<param name="thresh" type="integer" value="5" />')
        return IntegerToolParameter(self, elem)

    def params_from_strings(self, params, app, ignore_errors=False):
        return params_from_strings(self.inputs, params, app, ignore_errors)

    @property
    def config_file(self):
        return "<fake tool>"

    @property
    def template_macro_params(self):
        return {}

    @property
    def inputs(self):
        return self._params

    def set_params(self, params):
        self._params = params

    @property
    def outputs(self):
        return dict(
            output1=ToolOutput("output1"),
        )

    @property
    def tmp_directory_vars(self):
        return ["TMP"]

    @property
    def config_files(self):
        return self._config_files

    @property
    def command(self):
        return self._command_line

    @property
    def interpreter(self):
        return None

    def handle_unvalidated_param_values(self, input_values, app):
        pass

    def build_param_dict(self, incoming, *args, **kwds):
        return incoming

    def call_hook(self, hook_name, *args, **kwargs):
        self.hooks_called.append(hook_name)

    def exec_before_job(self, *args, **kwargs):
        pass





from galaxy2janis.logs import logging
import regex as re

from .. import utils

from galaxy2janis import expressions
from galaxy2janis.expressions.patterns import (
    FUNCTION_CALL_FMT1,
    FUNCTION_CALL_FMT2,
    BACKTICK_SECTION,
    QUOTED_SECTION,
    GX_DYNAMIC_KEYWORDS,
    # GX_STATIC_KEYWORDS,
)


def replace_function_calls(cmdstr: str) -> str:
    cmdlines = utils.split_lines(cmdstr)
    out: list[str] = []
    for line in cmdlines:
        matches = expressions.get_matches(line, FUNCTION_CALL_FMT1)
        matches += expressions.get_matches(line, FUNCTION_CALL_FMT2)
        for match in matches:
            logging.has_cheetah_function()
            old_section = match[0]
            new_section = '__FUNCTION_CALL__'
            line = line.replace(old_section, new_section)
        out.append(line)
    return utils.join_lines(out)

def replace_backticks(cmdstr: str) -> str:
    matches = expressions.get_matches(cmdstr, BACKTICK_SECTION)
    for match in matches:
        logging.has_backtick_statement()
        old_section = match[0]
        new_section = '__BACKTICK_SHELL_STATEMENT__'
        cmdstr = cmdstr.replace(old_section, new_section)
    return cmdstr

def interpret_raw(cmdstr: str) -> str:
    return cmdstr.replace('\\', '')

def flatten_multiline_strings(cmdstr: str) -> str:
    matches = expressions.get_matches(cmdstr, QUOTED_SECTION)
    for match in matches:
        if '\n' in match[0]:
            logging.has_multiline_str()
            old_section = match[0]
            new_section = match[0].replace('\n', ' ')
            cmdstr = cmdstr.replace(old_section, new_section)
    return cmdstr

def translate_variable_markers(cmdstr: str) -> str:
    return cmdstr.replace("gxparam_", "$")

def standardise_variable_format(cmdstr: str) -> str:
    """standardises different forms of a galaxy variable ${var}, $var etc"""
    cmdlines: list[str] = utils.split_lines(cmdstr)
    outlines: list[str] = [remove_var_braces(line) for line in cmdlines]
    return utils.join_lines(outlines)

def remove_var_braces(text: str) -> str:
    """
    modifies cmd word to ensure the $var format is present, rather than ${var}
    takes a safe approach using regex and resolving all vars one by one
    """
    return text
    # if text == '':
    #     return text
    # matches = scanners.get_variables_fmt2(text)
    # if len(matches) > 0:
    #     m = matches[0]
    #     # this is cursed but trust me it removes the curly braces for the match span
    #     old_segment = text[m.start(): m.end()]
    #     new_segment = old_segment[0] + old_segment[2: -1]
    #     new_segment = new_segment.replace(' ', '')
    #     text = text.replace(old_segment, new_segment)
    #     text = remove_var_braces(text)
    # return text  

def remove_empty_quotes(cmdstr: str) -> str:
    cmdstr = cmdstr.replace('""', '')
    cmdstr = cmdstr.replace("''", '')
    return cmdstr

def simplify_sh_constructs(cmdstr: str) -> str:
    """
    this function standardises the different equivalent 
    forms of linux operations into a single common form
    """
    cmdstr = cmdstr.replace("&amp;", "&")
    cmdstr = cmdstr.replace("&lt;", "<")
    cmdstr = cmdstr.replace("&gt;", ">")
    cmdstr = cmdstr.replace("|&", "2>&1 |")
    cmdstr = cmdstr.replace("| tee", "|tee")
    cmdstr = cmdstr.replace("1>", ">")
    return cmdstr 

def simplify_galaxy_dynamic_vars(cmdstr: str) -> str:
    """  ${GALAXY_SLOTS:-2} -> 2   etc """
    matches = expressions.get_matches(cmdstr, GX_DYNAMIC_KEYWORDS)
    for match in matches:
        cmdstr = cmdstr.replace(match[0], match.group(1)) 
    return cmdstr

def remove_cheetah_comments(cmdstr: str) -> str:
    """
    removes cheetah comments from shellparser lines
    comments can be whole line, or part way through
    """
    cmdlines: list[str] = utils.split_lines(cmdstr)
    outlines: list[str] = []

    for line in cmdlines:
        comment_start, _ = expressions.find_unquoted(line, '##')
        if comment_start != -1:
            # override line with comment removed
            line = line[:comment_start].strip()
        # make sure we didnt trim a full line comment and now its an empty string
        if line != '':
            outlines.append(line)
    return utils.join_lines(outlines)







from typing import Optional, Any

from .JanisDatatype import JanisDatatype
from .strategies import strategy_map
from .conversion import janis_to_core
from .conversion import select_primary_core_type
from .register import register


def get(entity: Any, entity_type: Optional[str]=None, source: str='galaxy') -> JanisDatatype:
    etype = entity_type if entity_type else entity.__class__.__name__
    strategy = strategy_map[etype] 
    jtypes = strategy().get(entity)
    core_types = janis_to_core(jtypes)
    return select_primary_core_type(core_types)

def populate() -> None:
    register.populate()






from typing import Optional, Tuple

import re

from .. import utils

from galaxy2janis import expressions
from galaxy2janis.expressions.patterns import (
    CHEETAH_SET,
    LINUX_LN,
    LINUX_MV,
    LINUX_CP
)


def resolve_aliases(cmdstr: str) -> str:
    resolver = AliasResolver()
    return resolver.resolve(cmdstr)

def extract_alias(line: str) -> Tuple[Optional[str], Optional[str]]:
    source, dest = None, None
    for func in [
        get_set,
        get_symlink, 
        #get_move, 
        #get_copy, 
    ]:
        source, dest = func(line)
        if source and dest:
            break
    return source, dest

def is_alias_line(line: str) -> bool:
    source, dest = extract_alias(line)
    if source and dest:
        return True
    return False

def get_set(line: str) -> Tuple[Optional[str], Optional[str]]:
    matches = expressions.get_matches(line, CHEETAH_SET)
    if matches:
        m = matches[0]
        return m.group(1), m.group(2)
    return None, None

def get_symlink(line: str) -> Tuple[Optional[str], Optional[str]]:
    matches = expressions.get_matches(line, LINUX_LN)
    if matches:
        m = matches[0]
        return m.group(2), m.group(1)
    return None, None

def get_copy(line: str) -> Tuple[Optional[str], Optional[str]]:
    matches = expressions.get_matches(line, LINUX_CP)
    if matches:
        m = matches[0]
        return m.group(2), m.group(1)
    return None, None

def get_move(line: str) -> Tuple[Optional[str], Optional[str]]:
    matches = expressions.get_matches(line, LINUX_MV)
    if matches:
        m = matches[0]
        return m.group(1), m.group(2)
    return None, None




# class because it may change in future. 
# want to keep the .get() and .add() methods stable
class AliasRegister:
    def __init__(self):
        self.aliases: dict[str, list[str]] = {}
    
    def get(self, source: str) -> Optional[str]:
        if source in self.aliases:
            destinations = self.aliases[source]
            if len(destinations) == 1:  # won't return if ambiguity regarding dest
                return destinations[0]
        return None
    
    def add(self, source: str, dest: str) -> None:
        if source not in self.aliases:
            self.aliases[source] = []    
        self.aliases[source].append(dest)


class AliasResolver:
    def __init__(self):
        self.register: AliasRegister = AliasRegister()

    def resolve(self, cmdstr: str) -> str:
        self.register_aliases(cmdstr)
        return self.resolve_aliases(cmdstr)

    def register_aliases(self, cmdstr: str) -> None:
        lines = utils.split_lines(cmdstr)
        for line in lines:
            self.register_line(line)

    def register_line(self, line: str) -> None:
        source, dest = extract_alias(line)
        if source and dest and source != dest:
            self.register.add(source, dest)

    def resolve_aliases(self, cmdstr: str) -> str:
        resolved_lines: list[str] = []
        for line in utils.split_lines_whitespace(cmdstr):
            line = self.resolve_line(line)
            resolved_lines.append(line)
        cmdstr = utils.join_lines(resolved_lines)
        return cmdstr

    def resolve_line(self, line: str) -> str:
        if not is_alias_line(line):
            for source in self.register.aliases:
                dest = self.register.get(source)
                if dest:
                    # ensure ends/starts with whitespace or quotes or forwardslash or is curly brackets
                    matches = self.get_line_matches(source, line)
                    for match in matches:
                        line = self.perform_replacement(match, dest, line)
        return line

    def get_line_matches(self, source: str, line: str) -> list[re.Match[str]]:
        pattern = re.escape(source)
        raw_matches = re.finditer(pattern, line)
        delims = ['"', "'", ' ', '}']
        matches: list[re.Match[str]] = []

        for m in raw_matches:
            if m.end() == len(line):
                matches.append(m)
            elif line[m.end()] in delims:
                matches.append(m)
            elif m.end()+1 < len(line) and line[m.end()+1] in delims:
                matches.append(m)
        
        return matches

    def perform_replacement(self, match: re.Match[str], dest: str, line: str) -> str:
        line = line[:match.start()] + dest + line[match.end():]
        return line
        
        




from galaxy2janis import settings
from galaxy2janis import paths
import shutil
from galaxy2janis import expressions

from galaxy2janis.entities.workflow import Workflow
from galaxy2janis.entities.workflow import WorkflowStep
from galaxy2janis.entities.workflow.input import WorkflowInput
from galaxy2janis.entities.workflow.step.inputs import InputValue, WorkflowInputInputValue
from galaxy2janis.datatypes.core import file_t

from ...command.components import InputComponent
from ...command.components import Positional
from ...command.components import Option
from ...wrappers import fetch_wrapper


def handle_tool_script_inputs(janis: Workflow) -> None:
    # sets tool input values as default
    # tool components which accepts scripts from the galaxy wrapper $__tool_directory
    for j_step in janis.steps:
        settings.tool.set(wrapper=j_step.metadata.wrapper)
        handle_step(janis, j_step)

def handle_step(janis: Workflow, j_step: WorkflowStep) -> None:
    for component in get_linkable_components(j_step):
        if component_is_script(component):
            # get filepath for script
            galaxy_path: str = component.values.unique[0] # type: ignore
            local_path = galaxy_path.replace('$__tool_directory__/', '')
            dest = paths.script(local_path)
            src = get_wrapper_script_path(local_path, j_step)

            # copy the script to the destination folder
            shutil.copy2(src, dest)
            
            # create worklow input & add to workflow
            winp = create_workflow_input(j_step, component, dest)
            janis.add_input(winp)
            
            # create step value & add to step 
            invalue = create_workflow_value(component, winp)
            j_step.inputs.add(invalue)

def component_is_script(component: InputComponent) -> bool:
    if isinstance(component, Positional) or isinstance(component, Option):
        if len(component.values.unique) == 1:
            if expressions.is_script(component.values.unique[0]):
                return True
    return False

def get_wrapper_script_path(local_path: str, step: WorkflowStep) -> str:
    wrapper = step.metadata.wrapper
    wrapper_path = fetch_wrapper(
        owner= wrapper.owner,
        repo= wrapper.repo,
        revision= wrapper.revision,
        tool_id= wrapper.tool_id
    )
    wrapper_dir = wrapper_path.rsplit('/', 1)[0]
    return f'{wrapper_dir}/{local_path}'

def get_linkable_components(j_step: WorkflowStep) -> list[InputComponent]:
    out: list[InputComponent] = []
    for component in j_step.tool.inputs:
        if not j_step.inputs.get(component.uuid):
            out.append(component)
    return out

def create_workflow_input(j_step: WorkflowStep, j_target: InputComponent, filepath: str) -> WorkflowInput:
    step_tag = j_step.tag
    input_tag = j_target.tag
    return WorkflowInput(
        _name=f'{step_tag}_{input_tag}',
        array=j_target.array,
        is_runtime=True,
        datatype=file_t,
        optional=j_target.optional,
        value=filepath
    )

def create_workflow_value(j_target: InputComponent, winp: WorkflowInput) -> InputValue:
    return WorkflowInputInputValue(
        component=j_target,
        input_uuid=winp.uuid,
        is_runtime=True
    )





from typing import Any, Optional, Protocol, Tuple

from galaxy2janis.gx.command.components import CommandComponent
from galaxy2janis.gx.command.components import InputComponent
from galaxy2janis.gx.command.components import Option

from galaxy2janis.datatypes import JanisDatatype



def format_docstring(component: CommandComponent) -> Optional[str]:
    raw_doc = component.docstring
    if raw_doc:
        return raw_doc.replace('"', "'")
    return None

def get_wrapped_default(component: InputComponent) -> Optional[str]:
    default = component.default_value
    if isinstance(component, Option) and default is not None:
        if isinstance(default, str) and '$' in default:
            default = None
    if should_quote(default, component):
        return f'"{default}"'
    return default

def format_imports(imports: list[Tuple[str, str]]) -> str:
    out_str: str = ''
    for x in imports:
        out_str += f'from {x[0]} import {x[1]}\n'
    return out_str


class TypeFormattable(Protocol):
    @property
    def optional(self) -> bool:
        ...

    @property
    def array(self) -> bool:
        ...
    
    @property
    def datatype(self) -> JanisDatatype:
        ...
    

def should_quote(default: Any, component: TypeFormattable) -> bool:
    jtype = component.datatype
    if jtype.classname == 'Int' or jtype.classname == 'Float':
        return False
    elif default is None:
        return False
    return True

def datatype_permits_default(jtype: JanisDatatype) -> bool:
    # check this component's datatypes permit a default value
    types_with_allowed_default = ['String', 'Float', 'Int', 'Boolean', 'Double']
    if jtype.classname in types_with_allowed_default:
        return True
    return False

def format_typestr(entity: TypeFormattable, fmt: str='definition') -> str:
    if fmt == 'definition':
        return fmt_typestr_definition(entity)
    elif fmt == 'value':
        return fmt_typestr_value(entity)
    else:
        raise RuntimeError()

def fmt_typestr_definition(entity: TypeFormattable) -> str:
    jtype = entity.datatype
    # not array not optional
    if not entity.optional and not entity.array: 
        out_str = f'{jtype.classname}'
    
    # array and not optional
    elif not entity.optional and entity.array: 
        out_str = f'Array({jtype.classname})'
    
    # not array and optional
    elif entity.optional and not entity.array: 
        out_str = f'{jtype.classname}(optional=True)'
    
    # array and optional
    elif entity.optional and entity.array: 
        out_str = f'Array({jtype.classname}(), optional=True)'
    return out_str

    # if len(types) > 1:
    #     dtype = ', '.join([x.classname for x in types])
    #     dtype = "UnionType(" + dtype + ")"
    # else:
    #     dtype = types[0].classname

def fmt_typestr_value(entity: TypeFormattable) -> str:
    jtype = entity.datatype
    # not array not optional
    if not entity.optional and not entity.array: 
        out_str = f'{jtype.classname}'
    
    # array and not optional
    elif not entity.optional and entity.array: 
        out_str = f'{jtype.classname} ARRAY'
    
    # not array and optional
    elif entity.optional and not entity.array: 
        out_str = f'{jtype.classname} OPTIONAL'
    
    # array and optional
    elif entity.optional and entity.array: 
        out_str = f'{jtype.classname} ARRAY OPTIONAL'
    return out_str

    # if len(types) > 1:
    #     dtype = ', '.join([x.classname for x in types])
    #     dtype = "UnionType(" + dtype + ")"
    # else:
    #     dtype = types[0].classname

    








from typing import Iterable
from galaxy2janis.gx.wrappers.scraping.revisions import get_revisions_in_repo
from galaxy2janis.gx.wrappers.scraping.wrappers import Revision
from wrappers.WrapperCache import WrapperCache
import gx.wrappers.scraping.utils as utils


# SCRAPING
def scrape_single_repo(owner: str, repo: str) -> None:
    print(f'scraping toolshed for {repo} wrappers')
    do_revisions_scrape(owner, repo)
    do_wrappers_scrape(owner, repo)

def do_revisions_scrape(owner: str, repo: str) -> None:
    revisions_data = utils.load_data(utils.REVISION_DATA_PATH) # load
    revisions_data[owner][repo] = get_revisions_in_repo(owner, repo) # alter
    utils.save_data(utils.REVISION_DATA_PATH, revisions_data) # save

def do_wrappers_scrape(owner: str, repo: str) -> None:
    cache = WrapperCache() # load
    
    for revision in scrapable_revisions(owner, repo):
        wrappers = revision.get_wrappers()
        for wrapper in wrappers:
            cache.add(wrapper)  # save

def scrapable_revisions(owner: str, repo: str) -> Iterable[Revision]:
    revision_data = utils.load_data(utils.REVISION_DATA_PATH)
    for revision in revision_data[owner][repo]: # should never have key errors
        yield Revision(
            owner=owner,
            repo=repo,
            revision=revision['revision'],
            date_created=revision['date_created'],
        )







    




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from galaxy2janis.entities.tool.Tool import Tool
    from galaxy2janis.entities.workflow import Workflow
    from galaxy2janis.entities.workflow import WorkflowInput
    from galaxy2janis.entities.workflow import WorkflowStep
    from galaxy2janis.entities.workflow import StepOutput

    from galaxy2janis.gx.command.components import InputComponent
    from galaxy2janis.gx.command.components import OutputComponent


# split this file into mapping.workflow, mapping.tool?

def tool_input(galaxy_param_name: str, tool: Tool) -> Optional[InputComponent]:
    """given a galaxy param name, get the corresponding Tool InputComponent if exists"""
    for inp in tool.inputs:
        if inp.gxparam and inp.gxparam.name == galaxy_param_name:
            return inp

def tool_output(galaxy_param_name: str, tool: Tool) -> Optional[OutputComponent]:
    """given a galaxy param name, get the corresponding Tool OutputComponent if exists"""
    for out in tool.outputs:
        if out.gxparam and out.gxparam.name == galaxy_param_name:
            return out

def emitter(query_id: str, query_output: str, janis: Workflow, galaxy: dict[str, Any]) -> StepOutput | WorkflowInput:
    """given a galaxy step id and output name, get the corresponding janis StepOutput or WorkflowInput"""
    g_step = _get_galaxy_step(query_id, galaxy)
    if g_step['type'] == 'tool':
        j_step = step(g_step['id'], janis, galaxy)
        return _get_janis_step_output(query_output, j_step, g_step)
    else:
        return _get_janis_w_input(g_step['id'], janis, galaxy)

def step(query_id: str, janis: Workflow, galaxy: dict[str, Any]) -> WorkflowStep:
    """lookup the corresponding janis step for a galaxy step with id=query_id"""
    g_step = _get_galaxy_step(query_id, galaxy)
    if 'janis_uuid' not in g_step:
        raise RuntimeError('Galaxy step not linked to janis step')
    janis_step_uuid = g_step['janis_uuid']
    return _get_janis_step(janis_step_uuid, janis)

def _get_galaxy_step(query_id: str, galaxy: dict[str, Any]) -> dict[str, Any]:
    for step in galaxy['steps'].values():
        if step['id'] == query_id:
            return step
    raise RuntimeError(f'No galaxy step with id {query_id}')

def _get_janis_step(query_uuid: str, janis: Workflow) -> WorkflowStep:
    for step in janis.steps:
        if step.uuid == query_uuid:
            return step
    raise RuntimeError(f'No janis step with uuid {query_uuid}')

def _get_janis_step_output(query_output: str, j_step: WorkflowStep, g_step: dict[str, Any]) -> StepOutput:
    """sorry"""
    j_outs = j_step.outputs.list()
    g_outs = g_step['outputs']
    for out in g_outs:
        if out['name'] == query_output:
            for output in j_outs:
                if output.uuid == out['janis_uuid']:
                    return output
    raise RuntimeError(f'No janis step output for galaxy step output {query_output}')

def _get_janis_w_input(query_id: str, janis: Workflow, galaxy: dict[str, Any]) -> WorkflowInput:
    g_step = _get_galaxy_step(query_id, galaxy)
    for inp in janis.inputs:
        if inp.uuid == g_step['janis_uuid']:
            return inp
    raise RuntimeError(f'No janis workflow input with uuid {query_id}')








from galaxy2janis import settings

from galaxy2janis.utils import galaxy as utils


def tool_setup() -> None:
    validate_tool_settings()


### VALIDATION ###

def validate_tool_settings() -> None:
    # both local & remote params not given
    if not _has_xml() or not _valid_xml():
        raise RuntimeError('no valid xml file')

def _has_xml() -> bool:
    if settings.tool.tool_path:
        return True
    return False

def _valid_xml() -> bool:
    path = settings.tool.tool_path
    if utils.is_tool_xml(path):
        return True
    return False







from setuptools import setup, find_packages
from pathlib import Path

# read the contents of README.md file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()
packages = ["galaxy2janis"] + ["galaxy2janis." + p for p in sorted(find_packages("./galaxy2janis"))]

setup(
    name='galaxy2janis',
    version='0.1.1',
    license='MIT',
    author='Grace Hall',
    description='ingestion of galaxy tool wrappers (.xml) and workflows (.ga) into the janis language.',
    packages=packages,
    package_data={'galaxy2janis': [
        'data/*.json',
        'data/*.yaml',
    ]},
    install_requires=[
        'galaxy-app==21.9.0',
        'janis-pipelines==0.11.6',
        'biopython==1.79',
        'filelock==3.7.0',
    ],
    entry_points={"console_scripts": ["galaxy2janis=galaxy2janis.run:main"]},
    long_description=long_description,
    long_description_content_type="text/markdown",
)




from galaxy2janis import expressions
from galaxy2janis.expressions.patterns import COMPOUND_OPT

from ..tokens import Token
from ..tokens import TokenType
from ...gxtool.param.InputParam import BoolParam, SelectParam


NON_VALUE_TOKENTYPES = set([
    TokenType.FUNCTION_CALL, 
    TokenType.BACKTICK_SHELL_STATEMENT, 
    TokenType.LINUX_TEE, 
    TokenType.LINUX_REDIRECT,
    TokenType.LINUX_STREAM_MERGE,
    TokenType.END_STATEMENT,
    TokenType.EXCISION,
])

def is_bool_select(token: Token) -> bool:
    if token.ttype == TokenType.GX_INPUT:
        match token.gxparam:
            case BoolParam():
                return True
            case SelectParam():
                if len(token.gxparam.options) > 0:
                    return True
            case _:
                pass
    return False


# FLAG STUFF

def boolparam_then_gxparam(ctoken: Token, ntoken: Token) -> bool:
    # for when boolparam followed by a different gxparam
    if looks_like_a_flag(ctoken):
        if isinstance(ctoken.gxparam, BoolParam):
            if ctoken.gxparam and ntoken.gxparam:  # curr and next both have gxparams
                if ctoken.gxparam.name != ntoken.gxparam.name:  # ensure not same gxparam
                    return True
    return False

def flag_then_flag(ctoken: Token, ntoken: Token) -> bool:
    if looks_like_a_flag(ctoken) and looks_like_a_flag(ntoken):
        return True
    return False

def flag_then_null(ctoken: Token, ntoken: Token) -> bool:
    if looks_like_a_flag(ctoken) and ntoken.ttype in NON_VALUE_TOKENTYPES:
        return True
    return False

def looks_like_a_flag(token: Token) -> bool:
    allowed_prefix_types = [TokenType.STRING, TokenType.INTEGER]
    if token.ttype == TokenType.FORCED_PREFIX:
        return True
    if token.ttype in allowed_prefix_types and token.text.startswith('-'):
        return True
    return False

flag_conditions = [
    boolparam_then_gxparam,
    flag_then_flag,
    flag_then_null,
]

def is_flag(ctoken: Token, ntoken: Token) -> bool:
    for condition in flag_conditions:
        if condition(ctoken, ntoken):
            return True
    return False


# OPTION STUFF 


def kvlinker(ctoken: Token, ntoken: Token) -> bool:
    if ntoken.ttype == TokenType.KV_LINKER:
        return True 
    return False

def compound_option(ctoken: Token, ntoken: Token) -> bool:
    if looks_like_a_flag(ctoken):
        if has_compound_structure(ctoken) and not is_positional(ntoken):
            return True
    return False

def has_compound_structure(token: Token) -> bool:
    compound_opts = expressions.get_matches(token.text, COMPOUND_OPT)
    if compound_opts:
        match = compound_opts[0]
        value = match.group(2)
        if int(value) > 3:
            return True
    return False

def prefix_then_value(ctoken: Token, ntoken: Token) -> bool:
    if not is_flag(ctoken, ntoken):
        if looks_like_a_flag(ctoken) and is_positional(ntoken):
            return True
    return False

option_conditions = [
    kvlinker,
    compound_option,
    prefix_then_value
]

def is_option(ctoken: Token, ntoken: Token) -> bool:
    """
    happens 2nd after 'is_flag()'
    already know that its not a flag, so if the current token
    looks like a flag/option, it has to be an option. 
    """
    if not within_different_constructs(ctoken, ntoken):
        for condition in option_conditions:
            if condition(ctoken, ntoken):
                return True
    return False

def within_different_constructs(ctoken: Token, ntoken: Token) -> bool:
    # one token is in a construct, other is not
    if ctoken.construct and not ntoken.construct:
        return True
    elif not ctoken.construct and ntoken.construct:
        return True
    # both tokens are in constructs, but the construct is not the same
    elif ctoken.construct and ntoken.construct:
        if ctoken.construct.uuid != ntoken.construct.uuid:
            return True
    return False


# POSITIONAL STUFF

def is_positional(token: Token) -> bool:
    if not looks_like_a_flag(token):
        if token.ttype not in NON_VALUE_TOKENTYPES:
            return True
    return False








from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Optional

from galaxy2janis import tags
from galaxy2janis import datatypes

from uuid import uuid4
from galaxy2janis.gx.gxtool.param import Param


class OutputComponent(ABC):
    def __init__(self):
        self.uuid: str = str(uuid4())
        self.gxparam: Optional[Param] = None
        self.forced_optionality: Optional[bool] = None
        self.forced_array: Optional[bool] = None
 
    @property
    @abstractmethod
    def name(self) -> str:
        """
        returns a name for this component. created depending on what
        information is available to the component
        """
        ...
    
    @property
    def tag(self) -> str:
        return tags.get(self.uuid)
    
    @property
    def datatype(self) -> datatypes.JanisDatatype:
        return datatypes.get(self)

    @property
    @abstractmethod
    def optional(self) -> bool:
        """
        returns whether the component is optional or not.
        uses galaxy param information if available, otherwise uses the presence array. flags are always optional
        """
        ...

    @property
    def array(self) -> bool:
        if self.forced_array is not None:
            return self.forced_array
        elif self.gxparam:
            return self.gxparam.array
        return False

    @property
    @abstractmethod
    def docstring(self) -> Optional[str]:
        """
        gets helptext for the component. uses galaxy param if available,
        otherwise usually just presents the witnessed values as examples. 
        """
        ...








from enum import Enum, auto

class Stream(Enum):
    STDIN = auto()
    STDOUT = auto()
    STDERR = auto()
    BOTH = auto()


class StreamMerge:
    def __init__(self, text: str):
        self.source: Stream = self.extract_source(text)
        self.destination: Stream = self.extract_dest(text)

    def extract_source(self, text: str) -> Stream:
        if text[0] == '2':
            return Stream.STDERR
        return Stream.STDOUT

    def extract_dest(self, text: str) -> Stream:
        if text[-1] == '2':
            return Stream.STDERR
        return Stream.STDOUT





from typing import Any, Optional

from galaxy2janis.gx.wrappers import Wrapper
from galaxy2janis.gx.wrappers import fetch_wrapper
from galaxy2janis.utils.galaxy import get_xml_id

tool_path: str
tool_id: str
owner: Optional[str] = None
repo: Optional[str] = None
revision: Optional[str] = None


# properties
def xml_basename() -> str:
    return tool_path.rsplit('/', 1)[-1].split('.')[0]

def xml_dir() -> str:
    if '/' not in tool_path:
        return '.'
    return tool_path.rsplit('/', 1)[0]

def logfile_path() -> str:
    return f'{xml_basename()}.log'

def set(args: Optional[dict[str, Any]]=None, wrapper: Optional[Wrapper]=None) -> None:
    if not args and not wrapper:
        raise RuntimeError('supply either args or wrapper to update')
    if args:
        update_args(args)
    elif wrapper:
        update_wrapper(wrapper)

def update_args(args: dict[str, Any]) -> None:
    global tool_path
    global tool_id
    global owner
    global repo
    global revision

    if args['infile']:
        tool_path = args['infile']
        tool_id = get_xml_id(tool_path)
        owner = None
        repo = None
        revision = None
    
    if args['remote']:
        owner, repo, tool_id, revision_raw = args['remote'].split(',')
        revision = revision_raw.rsplit(':', 1)[-1] # incase numeric:revision
        assert(owner)
        assert(repo)
        assert(revision)
        tool_path = fetch_wrapper(owner, repo, revision, tool_id)

def update_wrapper(wrapper: Wrapper) -> None:
    global tool_path
    global tool_id
    global owner
    global repo
    global revision

    tool_id = wrapper.tool_id
    owner = wrapper.owner
    repo = wrapper.repo
    revision = wrapper.revision
    tool_path = fetch_wrapper(wrapper.owner, wrapper.repo, wrapper.revision, wrapper.tool_id)







from typing import Tuple

from galaxy2janis.entities.workflow.step.outputs import StepOutput

from ..TextRender import TextRender
from .. import ordering
from .. import formatting

from galaxy2janis import tags  # remove if possible


class WorkflowOutputText(TextRender):
    def __init__(self, entity: StepOutput):
        super().__init__()
        self.entity = entity

    @property
    def imports(self) -> list[Tuple[str, str]]:
        jtype = self.entity.tool_output.datatype
        imports: list[Tuple[str, str]] = []
        imports.append((jtype.import_path, jtype.classname))

        # TODO opportunity for decorator
        if self.entity.tool_output.array:
            imports.append(('janis_core', 'Array'))

        # TODO opportunity for decorator
        imports = list(set(imports))
        return ordering.order_imports(imports)

    def render(self) -> str:
        step_tag = tags.get(self.entity.step_uuid)
        out_tag = self.entity.tool_output.tag
        datatype_str = formatting.format_typestr(self.entity.tool_output)
        out_str: str = ''
        out_str += 'w.output(\n'
        out_str += f'\t"{step_tag}_{out_tag}",\n'
        out_str += f'\t{datatype_str},\n'
        out_str += f'\tsource=(w.{step_tag}, "{out_tag}")'
        # out_str += f',\n\toutput_folder="{output_folder}"' if output_folder else ''
        # out_str += f',\n\toutput_name="{output_name}"' if output_name else ''
        # out_str += f',\n\textension="{extension}"' if extension else ''
        # out_str += f',\n\tdoc="{doc}"' if doc else ''
        out_str += '\n)\n'
        return out_str




from typing import Any
import regex as re
import keyword
import builtins

#from galaxy2janis.logs import logging

python_keys = set(keyword.kwlist)
builtin_keys = set(dir(builtins))
janis_keys = set([
    "identifier",
    "tool",
    "scatter",
    "ignore_missing",
    "output",
    "outputs",
    "input",
    "inputs"
])
keywords = python_keys | builtin_keys | janis_keys


def replace_keywords(tag: str, entity: Any) -> str:
    if tag in keywords:
        tag = _append_datatype(tag, entity)
    return tag

def encode(tag: str) -> str:
    return fr'{tag}'

def numeric(tag: str, entity: Any) -> str:
    """
    if tag is just a number, prepend the component type. 
    """
    if tag.replace('.','',1).isnumeric():  
        tag = _prepend_component_type(entity.cmd_pos, entity)
    return tag

def numeric_start(tag: str, entity: Any) -> str:
    """
    if tag starts with a number, prepend the component type. 
    """
    if len(tag) > 0 and tag[0].isnumeric():
        tag = _prepend_component_type(tag, entity)
    return tag

def short_tag(tag: str, entity: Any) -> str:
    if len(tag) == 0:
        #logging.zero_length_tag()
        tag = _prepend_component_type(tag, entity)
        tag = f'{tag}ERROR'
    elif len(tag) == 1:
        tag = _prepend_component_type(tag, entity)
    return tag

def camelify(text: str) -> str:
    text = re.sub(r"(_|-)+", " ", text)
    words = text.split()
    words = [w.title() for w in words]
    words = [words[0].lower()] + words[1:]
    return ''.join(words)

def non_alphanumeric(tag: str, entity: Any) -> str:
    """
    to satisfy janis tag requirements
    to avoid janis reserved keywords
    """
    tag = tag.strip('\\/-${}')
    tag = tag.replace('-', '_')
    tag = tag.replace('*', '')
    tag = tag.replace('(', '')
    tag = tag.replace(')', '')
    tag = tag.replace(']', '')
    tag = tag.replace('[', '')
    tag = tag.replace('.', '_')
    tag = tag.replace(' ', '_')
    tag = tag.replace('|', '_')
    tag = tag.replace('/', '_')
    tag = tag.replace('\\', '_')
    tag = tag.replace('"', '')
    tag = tag.replace("'", '')
    tag = tag.replace("@", 'at')
    return tag

def _prepend_component_type(tag: str, entity: Any) -> str:
    entity_type: str = entity.__class__.__name__
    entity_type = entity_type.lower()
    return f'{entity_type}_{tag}'

def _append_datatype(tag: str, entity: Any) -> str:
    jtype = entity.datatype
    jclass = jtype.classname.title()
    if not tag.endswith(jclass): # don't add the dtype if its already been added
        tag = f"{tag}{jclass}"
    return tag

# def _strip_numerals(tag: str) -> str:  # ??? y
#     return tag.lstrip('0123456789')







from __future__ import annotations
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from galaxy2janis.entities.workflow import WorkflowStep
    from galaxy2janis.entities.workflow import Workflow

from galaxy2janis.entities.workflow import WorkflowInput
from galaxy2janis.entities.workflow import InputValue
from galaxy2janis.entities.workflow import WorkflowInputInputValue

from galaxy2janis.gx.command.components import InputComponent

from galaxy2janis import mapping
from galaxy2janis import datatypes
from galaxy2janis import settings


def handle_tool_runtime_inputs(janis: Workflow, galaxy: dict[str, Any]) -> None:
    """runtime user inputs in step 'inputs'"""
    ingestor = RuntimeInputIngestor(janis, galaxy)
    ingestor.ingest()


class RuntimeInputIngestor:
    def __init__(self, janis: Workflow, galaxy: dict[str, Any]):
        self.janis = janis
        self.galaxy = galaxy

    def ingest(self) -> None:
        for step in self.galaxy['steps'].values():
            if step['type'] == 'tool':
                self.ingest_runtime(step)

    def ingest_runtime(self, g_step: dict[str, Any]) -> None:
        j_step = mapping.step(g_step['id'], self.janis, self.galaxy)
        settings.tool.set(wrapper=j_step.metadata.wrapper)
        g_targets = [inp['name'] for inp in g_step['inputs']]
        
        for g_target in g_targets:
            g_target = g_target.replace('|', '.')
            j_target = mapping.tool_input(g_target, j_step.tool)

            if not self.already_ingested(g_target, g_step):
                if not self.already_assigned(j_target, j_step):
                    if j_target: # only care if can link. is this a TODO?
                        # create workflow input & add to workflow
                        winp = self.create_workflow_input(j_step, j_target)
                        self.janis.add_input(winp)
                        # create step value & add to step 
                        value = self.create_workflow_value(j_target, winp)
                        j_step.inputs.add(value)

    def create_workflow_input(self, j_step: WorkflowStep, j_target: InputComponent) -> WorkflowInput:
        step_tag = j_step.tag
        input_tag = j_target.tag
        return WorkflowInput(
            _name=f'{step_tag}_{input_tag}',
            array=j_target.array,
            is_runtime=True,
            datatype=datatypes.get(j_target),
            optional=j_target.optional
        )

    def create_workflow_value(self, j_target: InputComponent, winp: WorkflowInput) -> InputValue:
        return WorkflowInputInputValue(
            component=j_target,
            input_uuid=winp.uuid,
            is_runtime=True
        )

    def already_ingested(self, g_target: str, gstep: dict[str, Any]) -> bool:
        if g_target in gstep['input_connections']:
            return True
        return False
    
    def already_assigned(self, j_target: Optional[InputComponent], j_step: WorkflowStep) -> bool:
        if j_target:
            if j_step.inputs.get(j_target.uuid):
                return True
        return False







from typing import Any
from copy import deepcopy


def get_flattened_tool_state(gxstep: dict[str, Any]) -> dict[str, Any]:
    flattener = ToolStateFlattener()
    return flattener.flatten(gxstep)


class ToolStateFlattener:
    def __init__(self):
        self.flattened_tool_state: dict[str, Any] = {}

    def flatten(self, step: dict[str, Any]) -> dict[str, Any]:
        for name, value in step['tool_state'].items():
            self.explore_node(name, value, [])
        return self.flattened_tool_state

    def explore_node(self, name: str, value: Any, path: list[str]) -> Any:
        path_copy = deepcopy(path)
        if name == '__current_case__':
            pass
        elif value == {"__class__": "RuntimeValue"}:
            pass
            #self.add_to_flattened_tool_state(name, 'RuntimeValue', path_copy)
        elif value == {"__class__": "ConnectedValue"}:
            pass
        elif isinstance(value, dict):
            path_copy.append(name)
            for key, val in value.items():
                self.explore_node(key, val, path_copy)
        else:
            self.add_to_flattened_tool_state(name, value, path_copy)
    
    def add_to_flattened_tool_state(self, name: str, value: Any, path_copy: list[str]) -> None:
        if len(path_copy) > 0:
            full_name = f'{".".join(path_copy)}.{name}'
        else:
            full_name = name
        self.flattened_tool_state[full_name] = value







from galaxy2janis.entities.workflow.workflow import Workflow

from .optionality import update_components_optionality
from .arrays import update_components_array
from .datatypes import update_components_datatype


def update_component_knowledge(janis: Workflow) -> None:
    for step in janis.steps:
        for value in step.inputs.linked:
            update_components_optionality(value, janis)
            update_components_array(value, janis)
            update_components_datatype(value, janis)
        



from galaxy2janis.logs import logging

from galaxy2janis.startup import tool_setup
from galaxy2janis.gx.gxtool import load_xmltool

from galaxy2janis.gx.command import gen_command
from galaxy2janis.containers import fetch_container
from galaxy2janis.entities.tool.generate import gen_tool
from galaxy2janis.entities.tool import Tool

# TODO future 
# from galaxy2janis.gx.xmltool.tests import write_tests

"""
this file parses a single tool to janis
the steps involved are laid out in order
each step involves a single module
"""

def tool_mode() -> Tool:
    tool_setup()
    logging.msg_parsing_tool()
    xmltool = load_xmltool()
    command = gen_command(xmltool)
    container = fetch_container(xmltool.metadata.get_main_requirement())
    tool = gen_tool(xmltool, command, container)
    return tool







import os
import unittest

import containers
from gx.gxtool.requirements import CondaRequirement

QUERY1 = CondaRequirement(_name='abricate', _version='1.0.1')
QUERY1_EXPECTED_RESULT = 'quay.io/biocontainers/abricate:1.0.1--ha8f3691_1'

QUERY2 = CondaRequirement(_name='samtools', _version='1.15')
QUERY2_EXPECTED_RESULT = 'quay.io/biocontainers/samtools:1.15--h1170115_1'

QUERY3 = CondaRequirement(_name='cutadapt', _version='3.5')
QUERY3_EXPECTED_RESULT = 'quay.io/biocontainers/cutadapt:3.5--py36h91eb985_1'


class TestContainerFetching(unittest.TestCase):
    """
    tests biocontainer fetching given a tool id, tool version, and a Requirement()
    """

    def setUp(self) -> None:
        """
        creates a temp container cache for use. the tearDown() method will remove this after tests have run. 
        """
        self.temp_cache_dir = 'test/temp/container_cache.json'
        with open(self.temp_cache_dir, 'w') as fp:
            fp.write('{}')
    
    def test_1(self) -> None:
        result = containers.fetch_container(QUERY1)
        self.assertEquals(result, QUERY1_EXPECTED_RESULT)
    
    def test_2(self) -> None:
        result = containers.fetch_container(QUERY2)
        self.assertEquals(result, QUERY2_EXPECTED_RESULT)
    
    def test_3(self) -> None:
        result = containers.fetch_container(QUERY3)
        self.assertEquals(result, QUERY3_EXPECTED_RESULT)
    
    def tearDown(self) -> None:
        if os.path.exists(self.temp_cache_dir):
            os.remove(self.temp_cache_dir)
        







from typing import Any

from galaxy2janis.gx.gxtool import load_xmltool
from galaxy2janis.gx.gxtool.param import BoolParam
from galaxy2janis.gx.gxtool import XMLToolDefinition


def resolve_values(gxstep: dict[str, Any]) -> dict[str, Any]:
    xmltool = load_xmltool()
    resolver = StepValueResolver(gxstep, xmltool)
    return resolver.resolve()



class StepValueResolver:
    def __init__(self, gxstep: dict[str, Any], xmltool: XMLToolDefinition):
        self.gxstep = gxstep
        self.xmltool = xmltool
        self.tool_state: dict[str, Any] = {}

    def resolve(self) -> dict[str, Any]:
        for key, val in self.gxstep['tool_state'].items(): 
            self.tool_state[key] = self.resolve_value(key, val)
        return self.tool_state
    
    def resolve_value(self, gxvarname: str, val: Any) -> Any:
        param = self.xmltool.inputs.get(gxvarname)
        match param:
            case BoolParam():
                if val == 'false':
                    return param.falsevalue
                else:
                    return param.truevalue
            case _:
                return val









from galaxy2janis.gx.gxtool.requirements import Requirement

from ..Container import Container


image_presets = {
    'python': [
        {
            'image_type': 'docker',
            'repo': 'python',
            'tag': '3.10',
            'url': 'quay.io/biocontainers/python:3.10',
            '_timestamp': 'Tue, 1 Mar 2022 18:45:00 -0000',
        }
    ]
}

def get_images_preset(requirement: Requirement) -> list[Container]:
    out: list[Container] = []
    if requirement.name in image_presets:
        for details in image_presets[requirement.name]:
            out.append(Container(details))
    return out
